#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import datetime
import json
from decimal import Decimal


class JsonHelper:

    @staticmethod
    def to_json(o):
        return json.dumps(o)

    @staticmethod
    def to_json_pretty(o):
        return json.dumps(o, indent=2)

    @staticmethod
    def to_jsonnable(o):
        if o is None \
            or isinstance(o, str) \
            or isinstance(o, int) \
            or isinstance(o, float) \
            or isinstance(o, bool):
            return o
        if isinstance(o, dict):
            for key, value in o.items():
                update = False
                if not isinstance(key, str):
                    del o[key]
                    key = str(key)
                    update = True
                jsonnable_value = JsonHelper.to_jsonnable(value)
                if value is not jsonnable_value:
                    value = jsonnable_value
                    update = True
                if update:
                    o[key] = value
            return o
        if isinstance(o, list):
            for i in range(len(o)):
                element = o[i]
                jsonnable_element = JsonHelper.to_jsonnable(element)
                if element is not jsonnable_element:
                    o[i] = jsonnable_element
            return o
        if isinstance(o, Decimal):
            return float(o)
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        if isinstance(o, datetime.date):
            return o.strftime('%Y-%m-%d')
        if isinstance(o, datetime.time):
            return o.strftime('%H:%M:%S')
        raise RuntimeError(f"Don't know how to jsonize {o} ({type(o)})")



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass
from deprecated import deprecated
from sodasql.scan.test import Test


@dataclass
class ScanError:
    message: str
    exception: Exception = None

    def __str__(self) -> str:
        return f'[{self.get_type()}] {self.get_message()}'

    def to_dict(self) -> dict:
        json = {
            'type': self.get_type(),
            'message': self.get_message()
        }
        if self.exception is not None:
            json['exception'] = str(self.exception)

            if hasattr(self.exception, "error_code"):
                json['errorCode'] = self.exception.error_code
        return json

    @deprecated(version='2.1.0b19', reason='This function is deprecated, please use to_dict')
    def to_json(self):
        return self.to_dict()

    def get_type(self) -> str:
        return 'error'

    def get_message(self) -> str:
        return self.message


@dataclass
class TestExecutionScanError(ScanError):
    test: Test = None

    def get_type(self) -> str:
        return 'test_execution_error'


@dataclass
class SodaCloudScanError(ScanError):
    def get_type(self) -> str:
        return 'soda_cloud_error'


@dataclass
class WarehouseAuthenticationScanError(ScanError):
    def get_type(self) -> str:
        return 'warehouse_authentication_error'


@dataclass
class WarehouseConnectionScanError(ScanError):
    def get_type(self) -> str:
        return 'warehouse_connection_error'



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json
from dataclasses import dataclass
from deprecated import deprecated
from typing import Optional

from sodasql.common.json_helper import JsonHelper
from sodasql.scan.test import Test


@dataclass
class TestResult:
    test: Test
    passed: bool
    skipped: bool
    values: Optional[dict] = None
    error: Optional[Exception] = None
    group_values: Optional[dict] = None

    def __str__(self):
        if self.passed:
            status_str = 'passed'
        elif self.skipped:
            status_str = 'skipped'
        else:
            status_str = 'failed'
        return (f'Test {self.test.title} {status_str}' +
                (f" with group values {self.group_values}" if self.group_values else '') +
                f' with measurements {json.dumps(JsonHelper.to_jsonnable(self.values))}')

    def to_dict(self) -> dict:
        if not self.test or not self.test.expression:
            return {
                'error': 'Invalid test result'
            }

        test_result_json = {
            'id': self.test.id,
            'title': self.test.title,
            'description': self.test.title,  # for backwards compatibility
            'expression': self.test.expression
        }

        if self.test.column:
            test_result_json['columnName'] = self.test.column

        if self.error:
            test_result_json['error'] = str(self.error)
        else:
            test_result_json['passed'] = self.passed
            test_result_json['skipped'] = self.skipped
            test_result_json['values'] = JsonHelper.to_jsonnable(self.values)

        if self.group_values:
            test_result_json['groupValues'] = JsonHelper.to_jsonnable(self.group_values)

        return test_result_json

    @deprecated(version='2.1.0b19', reason='This function is deprecated, please use to_dict')
    def to_json(self):
        return self.to_dict()



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass
from deprecated import deprecated
from sodasql.scan.group_value import GroupValue
from sodasql.scan.measurement import Measurement


@dataclass
class MonitorMeasurement(Measurement):
    metric_id: str = None
    sql: str = None
    query_milliseconds: int = None

    @classmethod
    def from_dict(cls, dictionary: dict) -> 'MonitorMeasurement':
        assert isinstance(dictionary, dict)
        return MonitorMeasurement(
            metric_id=dictionary.get('metricId'),
            metric=dictionary.get('metricType'),
            sql=dictionary.get('sql'),
            column_name=dictionary.get('columnName'),
            value=dictionary.get('value'),
            group_values=GroupValue.from_json_list(dictionary.get('groupValues')),
            query_milliseconds=dictionary.get('queryMilliseconds'))

    @classmethod
    @deprecated(version='2.1.0b19', reason='This function is deprecated, please use to_dict')
    def from_json(cls, dictionary: dict):
        cls.from_dict(cls, dictionary)

    def to_dict(self) -> dict:
        dictionary = super().to_dict()
        dictionary['metricId'] = self.metric_id
        dictionary['sql'] = self.sql
        dictionary['queryMilliseconds'] = self.query_milliseconds
        return dictionary

    @deprecated(version='2.1.0b19', reason='This function is deprecated, please use to_dict')
    def to_json(self):
        return self.to_dict()



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
import re
import tempfile
from datetime import datetime
from typing import List

from deprecated import deprecated

from sodasql.scan.metric import Metric
from sodasql.scan.samples_yml import SamplesYml
from sodasql.scan.scan import Scan
from sodasql.scan.scan_column import ScanColumn
from sodasql.scan.test_result import TestResult

logger = logging.getLogger(__name__)


class Sampler:

    def __init__(self, scan: Scan):
        self.scan = scan
        self.scan_reference = self.scan.scan_reference
        self.scan_folder_name = (f'{self._fileify(self.scan.warehouse.name)}' +
                                 f'-{self._fileify(self.scan.scan_yml.table_name)}' +
                                 (f'-{self._fileify(self.scan.time)}' if isinstance(self.scan.time, str) else '') +
                                 (f'-{self.scan.time.strftime("%Y%m%d%H%M%S")}' if isinstance(self.scan.time,
                                                                                              datetime) else '') +
                                 f'-{datetime.now().strftime("%Y%m%d%H%M%S")}')

    def _fileify(self, name: str):
        return re.sub(r'[^A-Za-z0-9_.]*', '', name).lower()

    def get_samples(self, samples_yml: SamplesYml, measurement, test_results: List[TestResult]):
        """
        Builds and executes SQL to save a sample for the given measurement
        Args:
            samples_yml: For column metric samples, samples_yml will have the defaults from the scan samples
                         already resolved.
        """
        column_name = measurement.column_name
        scan_column: ScanColumn = self.scan.scan_columns[column_name.lower()] if column_name else None
        tests = [test_result.test for test_result in test_results]

        # rows_total is the total number of rows from which the sample is taken
        rows_total = measurement.value
        sample_name = None
        where_clause = None
        tablesample = None
        limit = None
        test_ids = None

        if measurement.metric == Metric.ROW_COUNT:
            sample_name = 'dataset'
            tablesample = samples_yml.table_tablesample
            limit = samples_yml.table_limit
        elif measurement.metric == Metric.MISSING_COUNT:
            sample_name = 'missing'
            where_clause = scan_column.missing_condition
            tablesample = samples_yml.failed_tablesample
            limit = samples_yml.failed_limit

            def is_missing_test(test):
                return test.metrics == [Metric.MISSING_COUNT] or test.metrics == [Metric.MISSING_PERCENTAGE]

            test_ids = [test.id for test in tests if is_missing_test(test) and test.column == column_name]
        elif measurement.metric == Metric.VALUES_COUNT:
            sample_name = 'values'
            where_clause = f'NOT ({scan_column.missing_condition})'
            tablesample = samples_yml.passed_tablesample
            limit = samples_yml.passed_limit
        elif measurement.metric == Metric.INVALID_COUNT:
            sample_name = 'invalid'
            where_clause = f'NOT ({scan_column.missing_condition}) AND NOT ({scan_column.valid_condition})'
            tablesample = samples_yml.failed_tablesample
            limit = samples_yml.failed_limit

            def is_invalid_test(test):
                return test.metrics == [Metric.INVALID_COUNT] or test.metrics == [Metric.INVALID_PERCENTAGE]

            test_ids = [test.id for test in tests if is_invalid_test(test) and test.column == column_name]
        elif measurement.metric == Metric.VALID_COUNT:
            sample_name = 'valid'
            where_clause = f'NOT ({scan_column.missing_condition}) AND ({scan_column.valid_condition})'
            tablesample = samples_yml.passed_tablesample
            limit = samples_yml.passed_limit

        sql = (f"SELECT * \n"
               f"FROM {self.scan.qualified_table_name}")

        if where_clause:
            sql += f" \nWHERE {where_clause} "

        if tablesample is not None:
            sql += f" \nTABLESAMPLE {tablesample};"
        else:
            sql += f" \nLIMIT {limit};"

        sample_description = (f'{self.scan.scan_yml.table_name}.sample'
                              if sample_name == 'dataset' else f'{self.scan.scan_yml.table_name}.{column_name}.{sample_name}')

        logger.debug(f'Sending sample {sample_description}')

        sample_columns, sample_rows, total_row_count = self._get_query_results_with_limit(sql, limit or 5)

        return {'sample_name': sample_name,
                'column_name': column_name,
                'test_ids': test_ids,
                'sample_columns': sample_columns,
                'sample_rows': sample_rows,
                'sample_description': sample_description,
                'total_row_count': rows_total}

    def send_samples_to_soda_cloud(self, samples_context):

        sample_name = samples_context.get('sample_name')
        column_name = samples_context.get('column_name')
        test_ids = samples_context.get('test_ids')
        sample_columns = samples_context.get('sample_columns')
        sample_description = samples_context.get('sample_description')
        total_row_count = samples_context.get('total_row_count')
        sample_rows = samples_context.get('sample_rows')
        stored_row_count = len(sample_rows)

        scan_folder_name = (f'{self._fileify(self.scan.warehouse.name)}' +
                            f'-{self._fileify(self.scan.scan_yml.table_name)}' +
                            (f'-{self._fileify(self.scan.time)}' if isinstance(self.scan.time, str) else '') +
                            (f'-{self.scan.time.strftime("%Y%m%d%H%M%S")}' if isinstance(self.scan.time,
                                                                                         datetime) else '') +
                            f'-{datetime.now().strftime("%Y%m%d%H%M%S")}')
        try:

            with tempfile.TemporaryFile() as temp_file:
                for row in sample_rows:
                    temp_file.write(bytearray(json.dumps(row), 'utf-8'))
                    temp_file.write(b'\n')

                temp_file_size_in_bytes = temp_file.tell()
                temp_file.seek(0)

                file_path = (f'{scan_folder_name}/' +
                             (f'{self._fileify(column_name)}_' if column_name else '') +
                             f'{sample_name}.jsonl')

                file_id = self.scan.soda_server_client.scan_upload(self.scan.scan_reference,
                                                                   file_path,
                                                                   temp_file,
                                                                   temp_file_size_in_bytes)

                self.scan.soda_server_client.scan_file(
                    scan_reference=self.scan.scan_reference,
                    sample_type=sample_name + 'Sample',
                    stored=stored_row_count,
                    total=total_row_count,
                    source_columns=sample_columns,
                    file_id=file_id,
                    column_name=column_name,
                    test_ids=test_ids)

                logger.debug(f'Sent sample {sample_description} ({stored_row_count}/{total_row_count}) to Soda Cloud')
        except Exception as e:
            logger.exception(f'Soda cloud error: Could not upload samples: {e}')
            raise e

    def create_file_path_failed_rows_sql_metric(self, column_name: str, metric_name: str):
        return (f'{self.scan_folder_name}/' +
                (f'{self._fileify(column_name)}_' if column_name else '') +
                'failed_rows_' +
                f'{self._fileify(metric_name)}' +
                '.jsonl')

    def __serialize_file_upload_value(self, value):
        if value is None \
            or isinstance(value, str) \
            or isinstance(value, int) \
            or isinstance(value, float):
            return value
        return str(value)

    def __get_sample_columns(self, cursor):
        return [
            {'name': d[0],
             'type': self.scan.warehouse.dialect.get_type_name(d)}
            for d in cursor.description
        ]

    def _get_query_results_with_limit(self, sql: str, limit: int):
        cursor = self.scan.warehouse.connection.cursor()
        try:
            logger.debug(f'Executing SQL query: \n{sql}')
            start = datetime.now()
            cursor.execute(sql)
            sample_columns = self.__get_sample_columns(cursor)
            stored_rows = total_rows = 0
            row = cursor.fetchone()
            sample_rows = []
            while row is not None:
                sample_values = []
                if stored_rows < limit:
                    for i in range(0, len(row)):
                        sample_values.append(self.__serialize_file_upload_value(row[i]))
                    stored_rows += 1
                total_rows += 1
                sample_rows.append(sample_values)
                row = cursor.fetchone()

            delta = datetime.now() - start
            logger.debug(f'SQL took {str(delta)}')

        finally:
            cursor.close()

        return sample_columns, sample_rows, total_rows

    def save_sample_to_local_file_with_limit(self, sql, temp_file, limit: int):
        cursor = self.scan.warehouse.connection.cursor()
        try:
            logger.debug(f'Executing SQL query: \n{sql}')
            start = datetime.now()
            cursor.execute(sql)
            sample_columns = self.__get_sample_columns(cursor)

            stored_rows = total_rows = 0
            row = cursor.fetchone()
            while row is not None:
                if stored_rows < limit:
                    sample_values = []
                    for i in range(0, len(row)):
                        sample_values.append(self.__serialize_file_upload_value(row[i]))
                    temp_file.write(bytearray(json.dumps(sample_values), 'utf-8'))
                    temp_file.write(b'\n')
                    stored_rows += 1
                total_rows += 1
                row = cursor.fetchone()

            delta = datetime.now() - start
            logger.debug(f'SQL took {str(delta)}')

        finally:
            cursor.close()

        return stored_rows, sample_columns, total_rows



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import abc
import re
from datetime import date
import hashlib
import json
from numbers import Number
from typing import List, Optional
import importlib
import logging

from sodasql.exceptions.exceptions import WarehouseConnectionError, WarehouseAuthenticationError
from sodasql.scan.column_metadata import ColumnMetadata
from sodasql.scan.parser import Parser

KEY_WAREHOUSE_TYPE = 'type'
KEY_CONNECTION_TIMEOUT = 'connection_timeout_sec'
KEY_ACCOUNT_INFO_JSON_PATH = 'account_info_json_path'

ATHENA = 'athena'
BIGQUERY = 'bigquery'
HIVE = 'hive'
POSTGRES = 'postgres'
MYSQL = 'mysql'
REDSHIFT = 'redshift'
SNOWFLAKE = 'snowflake'
SQLSERVER = 'sqlserver'
SPARK = 'spark'

ALL_WAREHOUSE_TYPES = [ATHENA,
                       BIGQUERY,
                       HIVE,
                       POSTGRES,
                       MYSQL,
                       REDSHIFT,
                       SNOWFLAKE,
                       SQLSERVER,
                       SPARK]

logger = logging.getLogger(__name__)


class Dialect(metaclass=abc.ABCMeta):
    data_type_varchar_255 = "VARCHAR(255)"
    data_type_integer = "INTEGER"
    data_type_bigint = "BIGINT"
    data_type_decimal = "REAL"
    data_type_date = "DATE"

    def __init__(self, type: str):
        self.type = type

    @staticmethod
    def _import_class(module_name, class_name):
        _class_attr = None
        try:
            _module = importlib.import_module(module_name)
            _class_attr = getattr(_module, class_name)
        except ImportError:
            logger.error(f'Module {module_name} not found. Are you sure you installed appropriate warehouse package?')
        except AttributeError:
            logger.error(f'Class {class_name} not found in {module_name}, Are you sure you installed '
                         f'appropriate warehouse package?')

        return _class_attr

    @classmethod
    def create(cls, parser: Parser) -> Optional[Dialect]:
        _warehouse_class = None
        warehouse_type = parser.get_str_optional(KEY_WAREHOUSE_TYPE)
        if warehouse_type not in ALL_WAREHOUSE_TYPES:
            logger.error(
                f'Invalid warehouse type: {warehouse_type}, it must be one of {", ".join(ALL_WAREHOUSE_TYPES)}')
        else:
            if warehouse_type == ATHENA:
                _warehouse_class = Dialect._import_class('sodasql.dialects.athena_dialect', 'AthenaDialect')
            if warehouse_type == BIGQUERY:
                _warehouse_class = Dialect._import_class('sodasql.dialects.bigquery_dialect', 'BigQueryDialect')
            if warehouse_type == HIVE:
                _warehouse_class = Dialect._import_class('sodasql.dialects.hive_dialect', 'HiveDialect')
            if warehouse_type == POSTGRES:
                _warehouse_class = Dialect._import_class('sodasql.dialects.postgres_dialect', 'PostgresDialect')
            if warehouse_type == MYSQL:
                _warehouse_class = Dialect._import_class('sodasql.dialects.mysql_dialect', 'MySQLDialect')
            if warehouse_type == REDSHIFT:
                _warehouse_class = Dialect._import_class('sodasql.dialects.redshift_dialect', 'RedshiftDialect')
            if warehouse_type == SNOWFLAKE:
                _warehouse_class = Dialect._import_class('sodasql.dialects.snowflake_dialect', 'SnowflakeDialect')
            if warehouse_type == SQLSERVER:
                _warehouse_class = Dialect._import_class('sodasql.dialects.sqlserver_dialect', 'SQLServerDialect')
            if warehouse_type == SPARK:
                _warehouse_class = Dialect._import_class('sodasql.dialects.spark_dialect', 'SparkDialect')
        return _warehouse_class(parser)

    @classmethod
    def create_for_warehouse_type(cls, warehouse_type):
        from sodasql.scan.dialect_parser import DialectParser
        return cls.create(
            DialectParser(
                warehouse_connection_dict={
                    KEY_WAREHOUSE_TYPE: warehouse_type}))

    def default_connection_properties(self, params: dict):
        # to be overridden by subclass
        pass

    def safe_connection_data(self):
        """Return non-critically sensitive connection details.

        Useful for debugging.
        """
        # to be overridden by subclass
        pass

    def generate_hash_safe(self):
        """Generates a safe hash from non-sensitive connection details.

        Useful for debugging, identifying data sources anonymously and tracing.
        """
        data = self.safe_connection_data()

        return self.hash_data(data)

    def hash_data(self, data) -> str:
        """Hash provided data using a non-reversible hashing algorithm."""
        encoded = json.dumps(data, sort_keys=True).encode()
        return hashlib.sha256(encoded).hexdigest()

    def default_env_vars(self, params: dict):
        # to be overriden by subclass
        pass

    def query_table(self, table_name):
        query = f"""
            SELECT *
            FROM {table_name}
            LIMIT 1
        """
        return query

    def sql_test_connection(self):
        return "select 1"

    def sql_connection_test(self):
        return "select 1"

    @abc.abstractmethod
    def create_connection(self):
        pass

    @abc.abstractmethod
    def sql_columns_metadata_query(self, table_name: str) -> str:
        pass

    @abc.abstractmethod
    def sql_tables_metadata_query(self, limit: Optional[int] = None, filter: str = None):
        pass

    @abc.abstractmethod
    def is_text(self, column_type: str):
        pass

    @abc.abstractmethod
    def is_number(self, column_type: str):
        pass

    @abc.abstractmethod
    def is_time(self, column_type: str):
        pass

    def create_scan(self, *args, **kwargs):
        # Purpose of this method is to enable dialects to override and
        # customize the scan implementation
        from sodasql.scan.scan import Scan
        return Scan(*args, **kwargs)

    def is_supported(self, column_type: str):
        return (self.is_text(column_type)
                or self.is_number(column_type)
                or self.is_time(column_type))

    def sql_create_table(
        self,
        table_name: str,
        column_declarations: List[str]):
        columns_sql = ",\n  ".join(column_declarations)
        return f"CREATE TABLE " \
               f"{self.qualify_writable_table_name(table_name)} ( \n" \
               f"  {columns_sql} )"

    def sql_insert_into(self, table_name, rows: list):
        rows_sql = ',\n  '.join(rows)
        return (f'INSERT INTO '
                f"{self.qualify_writable_table_name(table_name)} VALUES \n"
                f"  {rows_sql}")

    def sql_drop_table(self, table_name):
        return f"DROP TABLE IF EXISTS {self.qualify_writable_table_name(table_name)}"

    def sql_expr_count_all(self) -> str:
        return 'COUNT(*)'

    def sql_expr_count_conditional(self, condition: str):
        return f'COUNT(CASE WHEN {condition} THEN 1 END)'

    def sql_expr_conditional(self, condition: str, expr: str):
        return f'CASE WHEN {condition} THEN {expr} END'

    def sql_expr_count(self, expr):
        return f'COUNT({expr})'

    def sql_expr_distinct(self, expr):
        return f'DISTINCT({expr})'

    def sql_expr_min(self, expr):
        return f'MIN({expr})'

    def sql_expr_length(self, expr):
        return f'LENGTH({expr})'

    def sql_expr_max(self, expr: str):
        return f'MAX({expr})'

    def sql_expr_avg(self, expr: str):
        return f'AVG({expr})'

    def sql_expr_sum(self, expr: str):
        return f'SUM({expr})'

    def sql_expr_variance(self, expr: str):
        return f'VARIANCE({expr})'

    def sql_expr_stddev(self, expr: str):
        return f'STDDEV({expr})'

    def sql_expr_regexp_like(self, expr: str, pattern: str):
        return f"REGEXP_LIKE({expr}, '{self.qualify_regex(pattern)}')"

    def sql_expr_limit(self, count):
        return f'LIMIT {count}'

    def sql_select_with_limit(self, table_name, count):
        return f'SELECT * FROM {table_name} LIMIT {count}'

    def sql_expr_list(self, column: ColumnMetadata, values: List[str]) -> str:
        if self.is_text(column.data_type):
            sql_values = [self.literal_string(value) for value in values]
        elif self.is_number(column.data_type):
            sql_values = [self.literal_number(value) for value in values]
        else:
            raise RuntimeError(
                f"Couldn't format list {str(values)} for column {str(column)}")
        return '(' + ','.join(sql_values) + ')'

    def sql_expr_cast_text_to_number(
        self, quoted_column_name, validity_format):
        if validity_format == 'number_whole':
            return f"CAST({quoted_column_name} AS {self.data_type_decimal})"
        not_number_pattern = self.qualify_regex(r"[^-[0-9]\.\,]")
        comma_pattern = self.qualify_regex(r"\\,")
        return f"CAST(REGEXP_REPLACE(REGEXP_REPLACE({quoted_column_name}, '{not_number_pattern}', ''), " \
               f"'{comma_pattern}', '.') AS {self.data_type_decimal})"

    def literal_number(self, value: Number):
        if value is None:
            return None
        return str(value)

    def literal_string(self, value: str):
        if value is None:
            return None
        return "'" + self.escape_metacharacters(value) + "'"

    def literal_list(self, l: list):
        if l is None:
            return None
        return '(' + (','.join([self.literal(e) for e in l])) + ')'

    def literal_date(self, date: date):
        date_string = date.strftime("%Y-%m-%d")
        return f"DATE '{date_string}'"

    def literal(self, o: object):
        if isinstance(o, Number):
            return self.literal_number(o)
        elif isinstance(o, str):
            return self.literal_string(o)
        elif isinstance(o, list) or isinstance(o, set) or isinstance(o, tuple):
            return self.literal_list(o)
        raise RuntimeError(
            f'Cannot convert type {type(o)} to a SQL literal: {o}')

    def qualify_table_name(self, table_name: str) -> str:
        return table_name

    def qualify_column_name(self, column_name: str):
        return column_name

    def qualify_writable_table_name(self, table_name: str) -> str:
        return table_name

    def qualify_regex(self, regex):
        return regex

    def qualify_string(self, value: str):
        return value

    @staticmethod
    def escape_metacharacters(value: str):
        return re.sub(r'(\\.)', r'\\\1', value)

    def sql_declare_string_column(self, column_name):
        return f"{column_name} {self.data_type_varchar_255}"

    def sql_declare_integer_column(self, column_name):
        return f"{column_name} {self.data_type_integer}"

    def sql_declare_decimal_column(self, column_name):
        return f"{column_name} {self.data_type_decimal}"

    def sql_declare_big_integer_column(self, column_name):
        return f"{column_name} {self.data_type_bigint}"

    def sql_expression(self, expression_dict: dict, **kwargs):
        if expression_dict is None:
            return None
        expr_type = expression_dict['type']
        if expr_type == 'number':
            sql = self.literal_number(expression_dict['value'])
        elif expr_type == 'string':
            sql = self.literal_string(expression_dict['value'])
        elif expr_type == 'time':
            if expression_dict['scanTime']:
                sql = self.literal(kwargs['scan_time'])
            else:
                raise RuntimeError('Unsupported time comparison! Only "scanTime" is supported')
        elif expr_type == 'columnValue':
            sql = expression_dict['columnName']
        elif expr_type == 'collection':
            # collection of string or number literals
            value = expression_dict['value']
            sql = self.literal_list(value)
        elif expr_type == 'equals':
            left = self.sql_expression(expression_dict['left'], **kwargs)
            right = self.sql_expression(expression_dict['right'], **kwargs)
            sql = self.sql_expr_equal(left, right)
        elif expr_type == 'lessThan':
            left = self.sql_expression(expression_dict['left'], **kwargs)
            right = self.sql_expression(expression_dict['right'], **kwargs)
            sql = self.sql_expr_less_than(left, right)
        elif expr_type == 'lessThanOrEqual':
            left = self.sql_expression(expression_dict['left'], **kwargs)
            right = self.sql_expression(expression_dict['right'], **kwargs)
            sql = self.sql_expr_less_than_or_equal(left, right)
        elif expr_type == 'greaterThan':
            left = self.sql_expression(expression_dict['left'], **kwargs)
            right = self.sql_expression(expression_dict['right'], **kwargs)
            sql = self.sql_expr_greater_than(left, right)
        elif expr_type == 'greaterThanOrEqual':
            left = self.sql_expression(expression_dict['left'], **kwargs)
            right = self.sql_expression(expression_dict['right'], **kwargs)
            sql = self.sql_expr_greater_than_or_equal(left, right)
        elif expr_type == 'between':
            clauses = []
            value = self.sql_expression(expression_dict['value'], **kwargs)
            gte = self.literal_number(expression_dict.get('gte'))
            gt = self.literal_number(expression_dict.get('gt'))
            lte = self.literal_number(expression_dict.get('lte'))
            lt = self.literal_number(expression_dict.get('lt'))
            if gte:
                clauses.append(self.sql_expr_less_than_or_equal(gte, value))
            elif gt:
                clauses.append(self.sql_expr_less_than(gt, value))
            if lte:
                clauses.append(self.sql_expr_less_than_or_equal(value, lte))
            elif lt:
                clauses.append(self.sql_expr_less_than(value, lt))
            sql = ' AND '.join(clauses)
        elif expr_type == 'in':
            left = self.sql_expression(expression_dict['left'], **kwargs)
            right = self.sql_expression(expression_dict['right'], **kwargs)
            sql = self.sql_expr_in(left, right)
        elif expr_type == 'contains':
            value = self.sql_expression(expression_dict['left'], **kwargs)
            substring = self.escape_metacharacters(
                expression_dict['right']['value'])
            sql = self.sql_expr_contains(value, substring)
        elif expr_type == 'startsWith':
            value = self.sql_expression(expression_dict['left'], **kwargs)
            substring = self.escape_metacharacters(
                expression_dict['right']['value'])
            sql = self.sql_expr_starts_with(value, substring)
        elif expr_type == 'endsWith':
            value = self.sql_expression(expression_dict['left'], **kwargs)
            substring = self.escape_metacharacters(
                expression_dict['right']['value'])
            sql = self.sql_expr_ends_with(value, substring)
        elif expr_type == 'not':
            sql = 'NOT (' + \
                  self.sql_expression(expression_dict['expression'], **kwargs) + ')'
        elif expr_type == 'and':
            sql = '(' + (') AND ('.join([self.sql_expression(e, **kwargs)
                                         for e in expression_dict['andExpressions']])) + ')'
        elif expr_type == 'or':
            sql = '(' + (') OR ('.join([self.sql_expression(e, **kwargs)
                                        for e in expression_dict['orExpressions']])) + ')'
        elif expr_type == 'null':
            sql = 'null'
        else:
            raise RuntimeError(f'Unsupported expression type: {expr_type}')
        return sql

    def sql_expr_equal(self, left, right):
        if right == 'null':
            return f'{left} IS NULL'
        else:
            return f'{left} = {right}'

    def sql_expr_less_than(self, left, right):
        return f'{left} < {right}'

    def sql_expr_less_than_or_equal(self, left, right):
        return f'{left} <= {right}'

    def sql_expr_greater_than(self, left, right):
        return f'{left} > {right}'

    def sql_expr_greater_than_or_equal(self, left, right):
        return f'{left} >= {right}'

    def sql_expr_in(self, left, right):
        return f'{left} IN {right}'

    def sql_expr_contains(self, value, substring):
        return value + " LIKE '%" + substring + "%'"

    def sql_expr_starts_with(self, value, substring):
        return value + " LIKE '" + substring + "%'"

    def sql_expr_ends_with(self, value, substring):
        return value + " LIKE '%" + substring + "'"

    def get_type_name(self, column_description):
        return str(column_description[1])

    def is_connection_error(self, exception):
        return False

    def is_authentication_error(self, exception):
        return False

    def try_to_raise_soda_sql_exception(
        self, exception: Exception) -> Exception:
        if self.is_connection_error(exception):
            raise WarehouseConnectionError(
                warehouse_type=self.type,
                original_exception=exception)
        elif self.is_authentication_error(exception):
            raise WarehouseAuthenticationError(
                warehouse_type=self.type, original_exception=exception)
        else:
            raise exception

    def sql_columns_metadata(self, table_name: str) -> List[tuple]:
        return []

    def validate_connection(self):
        # to be overriden by subclass
        pass



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging

import yaml

logger = logging.getLogger(__name__)


class YamlHelper:

    @staticmethod
    def parse_yaml(yaml_str: str, description: str = None):
        try:
            return yaml.load(yaml_str, Loader=yaml.SafeLoader)
        except Exception as e:
            logger.error(f'Parsing YAML failed: {str(e)}: ({description if description else yaml_str})xWW4')

    @staticmethod
    def validate_numeric_value(column_name, key, value):
        if value is None:
            logger.info(f'There is no value specified for {key} for column {column_name}')
        elif value is isinstance(value, int):
            logger.error(f'{column_name} could not be parsed: {key}-{value} is not of a numeric type.')
            raise Exception(f'{column_name} could not be parsed: {key}-{value} is not of a numeric type.')
        else:
            return value

    @staticmethod
    def validate_list_value(column_name, key, value):
        if value is None:
            logger.info(f'There is no value specified for {key} for column {column_name}')
        elif value is isinstance(value, list):
            logger.error(f'{column_name} could not be parsed: {key}-{value} is not of a list type.')
            raise Exception(f'{column_name} could not be parsed: {key}-{value} is not of a list type.')
        else:
            return value



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json
import logging
import os
import traceback
from collections import deque
from dataclasses import dataclass
from typing import Deque, List, Optional

import yaml
from sodasql.scan.test import Test

ERROR = 'error'
WARNING = 'warning'
INFO = 'info'


class ParseLog:
    def __init__(self, level: str, message: str, field: str):
        self.level = level
        self.message = message
        self.field = field

    def __str__(self):
        return f'[{self.level}] ({self.field}) {self.message}'

    def is_error_or_warning(self):
        return self.level == ERROR or self.level == WARNING

    def log(self):
        if self.level == ERROR:
            logging.error(self.message)
        elif self.level == WARNING:
            logging.warning(self.message)
        elif self.level == INFO:
            logging.info(self.message)
        else:
            logging.debug(self.message)


@dataclass
class ParseContext:
    name: str
    object: dict


class Parser:

    def __init__(self, description: str):
        self.description: str = description
        self.contexts: Deque[ParseContext] = deque()
        self.logs: List[ParseLog] = []

    def __str__(self):
        return '\n'.join([str(log) for log in self.logs])

    def _read_file_as_string(self, file_path: str):
        try:
            with open(file_path, encoding='utf-8') as f:
                return f.read()
        except Exception as e:
            self.error(f"Couldn't read file {file_path}: {str(e)}")

    def _parse_yaml_str(self, yaml_str):
        try:
            return yaml.load(yaml_str, Loader=yaml.SafeLoader)
        except Exception as e:
            self.error(f"Couldn't parse yaml in {self.description}: {str(e)}")

    def _push_context(self, object=None, name: Optional[object] = None):
        name = name if isinstance(name, str) or name is None else str(name)
        self.contexts.append(ParseContext(object=object, name=name))

    def _pop_context(self):
        return self.contexts.pop()

    def _get_context_description(self):
        return '.'.join([context.name for context in self.contexts if context.name])

    def error(self, message: str, field: str = None):
        return self.logs.append(ParseLog(ERROR, message, field))

    def warning(self, message: str, field: str = None):
        return self.logs.append(ParseLog(WARNING, message, field))

    def info(self, message, field: str = None):
        return self.logs.append(ParseLog(INFO, message, field))

    def log(self):
        for log in self.logs:
            log.log()

    def check_invalid_keys(self, valid_keys: List[str]):
        """
        Adds a warning for all invalid configured property names
        """
        context_iterable = self._get_current_context_object()
        for invalid_key in [configured_key for configured_key in context_iterable if configured_key not in valid_keys]:
            self.warning(f'Invalid key in {self._get_context_description()} : {invalid_key}', invalid_key)

    def has_warnings_or_errors(self):
        for log in self.logs:
            if log.is_error_or_warning():
                return True
        return False

    def assert_no_warnings_or_errors(self):
        if self.has_warnings_or_errors():
            raise AssertionError(f'{self.description} configuration errors: \n  '
                                 + ('\n  '.join([str(log) for log in self.logs])))

    def get_str_required(self, property_name: str):
        return self._get(property_name, str, True)

    def get_str_optional(self, property_name: str, default=None):
        return self._get(property_name, str, False, default)

    def get_str_required_env(self, property_name: str):
        return self._get(property_name, str, True)

    def get_credential(self, property_name: str):
        return self._get(property_name, str, False)

    def get_str_optional_env(self, property_name: str, default=None):
        return self._get(property_name, str, False, default)

    def get_int_required(self, property_name: str):
        return self._get(property_name, int, True)

    def get_int_optional(self, property_name: str, default=None):
        return self._get(property_name, int, False, default)

    def get_float_required(self, property_name: str):
        return self._get(property_name, float, True)

    def get_float_optional(self, property_name: str, default=None):
        return self._get(property_name, float, False, default)

    def get_bool_required(self, property_name: str):
        return self._get(property_name, bool, True)

    def get_bool_optional(self, property_name: str, default=None):
        return self._get(property_name, bool, False, default)

    def get_dict_required(self, property_name: str):
        return self._get(property_name, dict, True)

    def get_dict_optional(self, property_name: str, default=None):
        return self._get(property_name, dict, False, default)

    def get_list_required(self, property_name: str):
        return self._get(property_name, list, True)

    def get_list_optional(self, property_name: str, default=None):
        return self._get(property_name, list, False, default)

    def get_file_json_dict_required(self, property_name: str):
        file_str: str = self._read_file_as_string(property_name)
        try:
            return json.loads(file_str)
        except Exception as e:
            self.error(f"Couldn't parse json configuration {property_name} for {self.description}: {str(e)}",
                       property_name)

    def _get(self,
             property_name: str,
             return_type: type,
             is_required: bool,
             default=None,
             resolve_env: bool = True,
             env_required: bool = False):
        properties = self._get_current_context_object()
        if property_name in properties:
            value = properties.get(property_name)

            if resolve_env:
                if isinstance(value, str) \
                    and value.strip().startswith('env_var(') \
                    and value.strip().endswith(')'):
                    env_var_name = value.strip()[len('env_var('):-1]
                    value = os.getenv(env_var_name)
                    if not value:
                        self.info(f'Environment variable {env_var_name} is not set')
                elif env_required:
                    self.error(f'{self._get_context_description()}.{property_name} is considered a credential and '
                               f'must be passed as an environment variable eg: env_var(YOUR_ENVIRONMENT_VAR)')

            if value is None or isinstance(value, return_type):
                return value
            try:
                if return_type == int:
                    return int(value)
                if return_type == str:
                    return str(value)
                if return_type == float:
                    return float(value)
                raise ValueError(str(type(value)))
            except ValueError as e:
                self.error(f'Invalid {property_name}: Expected {str(return_type)}, but was {str(value)}: {str(e)}',
                           property_name)
                return default

        else:
            if is_required:
                self.error(f'Property {property_name} does not exist in {self._get_context_description()}',
                           property_name)
            return default

    def _get_current_context_object(self):
        return self.contexts[-1].object

    def parse_tests(self,
                    parent_dict: dict,
                    tests_key: str,
                    context_table_name: Optional[str] = None,
                    context_column_name: Optional[str] = None,
                    context_sql_metric_name: Optional[str] = None,
                    context_sql_metric_index: Optional[int] = None) -> List[Test]:
        tests: List[Test] = []

        test_ymls = parent_dict.get(tests_key)

        # Deprecated
        if isinstance(test_ymls, dict):
            test_ymls_list = []
            for test_name in test_ymls:
                test_ymls_list.append({
                    'name': test_name,
                    'expression': test_ymls[test_name]
                })
            test_ymls = test_ymls_list

        self._push_context(None, tests_key)
        try:
            if isinstance(test_ymls, list):
                for test_index in range(len(test_ymls)):
                    test_yml = test_ymls[test_index]
                    self._push_context(test_yml, str(test_index))
                    try:
                        test_name = None
                        test_title = None
                        test_expression = None

                        if isinstance(test_yml, str):
                            test_expression = test_yml
                        elif isinstance(test_yml, dict):
                            test_name = self.get_str_required('name')
                            test_title = self.get_str_optional('title')
                            test_expression = self.get_str_required('expression')
                        else:
                            self.error(f'Test must be a string or an object, but was {type(test_yml)}')

                        test = self.parse_test(test_index=test_index,
                                               test_name=test_name,
                                               test_title=test_title,
                                               test_expression=test_expression,
                                               context_table_name=context_table_name,
                                               context_column_name=context_column_name,
                                               context_sql_metric_name=context_sql_metric_name,
                                               context_sql_metric_index=context_sql_metric_index)
                        if test:
                            tests.append(test)
                    finally:
                        self._pop_context()

            elif test_ymls is not None:
                self.error(
                    f'Tests should be either a list of test expressions or an object of named test expressions: {test_ymls} ({str(type(test_ymls))})')
        finally:
            self._pop_context()

        return tests

    def parse_test(self,
                   test_expression: str,
                   test_name: str = None,
                   test_title: str = None,
                   test_index: int = None,
                   context_table_name: str = None,
                   context_column_name: str = None,
                   context_sql_metric_name: str = None,
                   context_sql_metric_index: int = None):

        if not test_expression:
            self.error('Test expression is required')
            return

        test_title = test_title if test_title else self.create_test_title(
            test_expression,
            test_name,
            test_index,
            context_column_name,
            context_sql_metric_name,
            context_sql_metric_index)

        test_id = self.create_test_id(
            test_expression,
            test_name,
            test_index,
            context_column_name,
            context_sql_metric_name,
            context_sql_metric_index)

        try:
            compiled_code = compile(test_expression, 'test', 'eval')

            metrics = None

            if context_table_name or context_column_name:
                names = compiled_code.co_names
                metrics = list(names)
                # non_metric_names = [name for name in names if name not in Metric.METRIC_TYPES]
                # if len(non_metric_names) != 0 or len(names) == 0:
                #     # Dunno yet if this should be info, warning or error.  So for now keeping it open.
                #     # SQL metric names and variables are not known until eval.
                #     self.info(f'At least one of the variables used in test ({set(names)}) '
                #               f'was not a valid metric type. Metric types: {Metric.METRIC_TYPES}, '
                #               f'Test: {test_description}')

            return Test(id=test_id,
                        title=test_title,
                        expression=test_expression,
                        metrics=metrics,
                        column=context_column_name)

        except SyntaxError:
            stacktrace_lines = traceback.format_exc().splitlines()
            self.error(f'Syntax error in test {test_title}:\n' +
                       ('\n'.join(stacktrace_lines[-3:])))

    def create_test_title(self,
                          test_expression,
                          test_name,
                          test_index,
                          context_column_name,
                          context_sql_metric_name,
                          context_sql_metric_index):
        parts = []
        if context_column_name:
            parts.append(f'column({context_column_name})')
        if context_sql_metric_index is not None:
            parts.append(f'sqlmetric({context_sql_metric_index})')
        elif context_sql_metric_name:
            parts.append(f'sqlmetric({context_sql_metric_name})')
        parts.append((test_name if test_name else f'test') + f'({test_expression})')
        return ' '.join(parts)

    def create_test_id(self,
                       test_expression,
                       test_name,
                       test_index,
                       context_column_name,
                       context_sql_metric_name,
                       context_sql_metric_index):
        # ORDERING AND CONTENTS ARE CRUCIAL
        # It's used to match tests with monitor ids on the Soda cloud platform
        test_id_dict = {}
        if context_column_name:
            test_id_dict['column'] = context_column_name

        if context_sql_metric_name:
            test_id_dict['sql_metric_name'] = context_sql_metric_name
        elif context_sql_metric_index is not None:
            test_id_dict['sql_metric_index'] = context_sql_metric_index

        if test_name:
            test_id_dict['test_name'] = test_name
        elif test_expression:
            test_id_dict['expression'] = test_expression
        else:
            test_id_dict['test_index'] = test_index

        return json.dumps(test_id_dict, separators=(',', ':'))



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass


@dataclass
class SamplesYml:
    table_limit: int
    table_tablesample: str
    failed_limit: int
    failed_tablesample: str
    passed_limit: int
    passed_tablesample: int

    def with_defaults(self, default_samples_yml):
        return SamplesYml(
            table_limit=self.table_limit if self.table_limit is not None else default_samples_yml.table_limit,
            table_tablesample=self.table_tablesample if self.table_tablesample is not None else default_samples_yml.table_tablesample,
            failed_limit=self.failed_limit if self.failed_limit is not None else default_samples_yml.failed_limit,
            failed_tablesample=self.failed_tablesample if self.failed_tablesample is not None else default_samples_yml.failed_tablesample,
            passed_limit=self.passed_limit if self.passed_limit is not None else default_samples_yml.passed_limit,
            passed_tablesample=self.passed_tablesample if self.passed_tablesample is not None else default_samples_yml.passed_tablesample,
        )

    def is_failed_enabled(self):
        return self.failed_limit or self.failed_tablesample

    def is_passed_enabled(self):
        return self.passed_limit or self.passed_tablesample

    def is_table_enabled(self):
        return self.table_limit is not None or self.table_tablesample is not None



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional

from sodasql.scan.dialect import Dialect
from sodasql.scan.group_value import GroupValue
from sodasql.scan.scan import Scan
from sodasql.scan.scan_column import ScanColumn

from sodasql.soda_server_client.monitor_measurement import MonitorMeasurement


class MonitorMetricType:
    ROW_COUNT = 'rowCount'
    MISSING_VALUES_COUNT = 'missingValuesCount'
    MISSING_VALUES_PERCENTAGE = 'missingValuesPercentage'
    VALID_VALUES_COUNT = 'validValuesCount'
    VALID_VALUES_PERCENTAGE = 'validValuesPercentage'
    INVALID_VALUES_COUNT = 'invalidValuesCount'
    INVALID_VALUES_PERCENTAGE = 'invalidValuesPercentage'
    MINIMUM_VALUE = 'minimumValue'
    MAXIMUM_VALUE = 'maximumValue'
    UNIQUE_VALUES_COUNT = 'uniqueValuesCount'
    DISTINCT_VALUES_COUNT = 'distinctValuesCount'
    UNIQUENESS_PERCENTAGE = 'uniquenessPercentage'
    MINIMUM_LENGTH = 'minimumLength'
    MAXIMUM_LENGTH = 'maximumLength'
    AVERAGE_LENGTH = 'averageLength'
    AVERAGE = 'average'
    STANDARD_DEVIATION = 'standardDeviation'
    SUM = 'sum'
    VARIANCE = 'variance'


@dataclass
class MonitorMetric:
    scan: Scan
    metric_id: str
    metric_type: str
    column_name: str
    group_by_column_names: List[str]
    metric_sql: str = None
    sql: str = None

    def build_sql(self,
                  qualified_group_column_names,
                  filter_condition,
                  qualified_table_name):

        scan_column: Optional[ScanColumn] = \
            self.scan.scan_columns.get(self.column_name.lower()) \
            if self.column_name and self.scan.scan_columns \
            else None

        dialect: Dialect = self.scan.warehouse.dialect

        metric_select_fields = []
        select_fields = []

        if qualified_group_column_names:
            metric_select_fields.extend(qualified_group_column_names)
            select_fields.extend(qualified_group_column_names)

        if self.metric_type == MonitorMetricType.ROW_COUNT:
            metric_select_fields.append(dialect.sql_expr_count_all())
        elif scan_column:
            select_fields.append(scan_column.qualified_column_name)
            if self.metric_type == MonitorMetricType.MISSING_VALUES_COUNT:
                metric_select_fields.append(dialect.sql_expr_count_conditional(scan_column.missing_condition))
            elif self.metric_type == MonitorMetricType.MISSING_VALUES_PERCENTAGE:
                metric_select_fields.append(dialect.sql_expr_count_all())
                metric_select_fields.append(dialect.sql_expr_count_conditional(scan_column.missing_condition))
            elif self.metric_type == MonitorMetricType.VALID_VALUES_COUNT:
                metric_select_fields.append(
                    dialect.sql_expr_count_conditional(scan_column.non_missing_and_valid_condition))
            elif self.metric_type in [MonitorMetricType.VALID_VALUES_PERCENTAGE,
                                      MonitorMetricType.INVALID_VALUES_COUNT,
                                      MonitorMetricType.INVALID_VALUES_PERCENTAGE]:
                metric_select_fields.append(dialect.sql_expr_count_conditional(scan_column.non_missing_condition))
                metric_select_fields.append(
                    dialect.sql_expr_count_conditional(scan_column.non_missing_and_valid_condition))
            elif self.metric_type == MonitorMetricType.UNIQUENESS_PERCENTAGE:
                metric_select_fields.append(
                    dialect.sql_expr_count_conditional(scan_column.non_missing_and_valid_condition))
                metric_select_fields.append(dialect.sql_expr_count(dialect.sql_expr_distinct(
                    dialect.sql_expr_conditional(scan_column.non_missing_and_valid_condition,
                                                 scan_column.qualified_column_name))))
            elif self.metric_type == MonitorMetricType.MINIMUM_VALUE:
                metric_select_fields.append(dialect.sql_expr_min(scan_column.numeric_expr))
            elif self.metric_type == MonitorMetricType.MAXIMUM_VALUE:
                metric_select_fields.append(dialect.sql_expr_max(scan_column.numeric_expr))
            elif self.metric_type == MonitorMetricType.AVERAGE:
                metric_select_fields.append(dialect.sql_expr_avg(scan_column.numeric_expr))
            elif self.metric_type == MonitorMetricType.SUM:
                metric_select_fields.append(dialect.sql_expr_sum(scan_column.numeric_expr))
            else:
                raise RuntimeError(f'Unsupported metric type: {self.metric_type}')

        metric_fields = ", \n       ".join(metric_select_fields)
        self.metric_sql = (f'SELECT {metric_fields} \n'
                           f'FROM {qualified_table_name}')

        self.sql = (f'SELECT * \n'
                    f'FROM {qualified_table_name}')

        if filter_condition:
            self.metric_sql += f' \nWHERE {filter_condition}'
            self.sql += f' \nWHERE {filter_condition}'

        if qualified_group_column_names:
            self.metric_sql += f' \nGROUP BY {", ".join(qualified_group_column_names)}'
            self.sql += f' \nGROUP BY {", ".join(qualified_group_column_names)}'

    def execute(self) -> MonitorMeasurement:
        if not self.group_by_column_names:
            start = datetime.now()
            row = self.scan.warehouse.sql_fetchone(self.metric_sql)
            query_milliseconds = int(((datetime.now() - start).total_seconds()) * 1000)
            value = self.get_value(row)
            return MonitorMeasurement(
                metric=self.metric_type,
                metric_id=self.metric_id,
                sql=self.metric_sql,
                column_name=self.column_name,
                value=value,
                query_milliseconds=query_milliseconds)
        else:
            start = datetime.now()
            rows = self.scan.warehouse.sql_fetchall(self.metric_sql)
            query_milliseconds = int(((datetime.now() - start).total_seconds()) * 1000)

            group_values = []
            for row in rows:
                results = list(row)
                group_columns_count = len(self.group_by_column_names)
                group = results[:group_columns_count]
                results = results[group_columns_count:]
                value = self.get_value(results)
                group_values.append(GroupValue(group=group, value=value))

            return MonitorMeasurement(
                metric=self.metric_type,
                metric_id=self.metric_id,
                sql=self.metric_sql,
                column_name=self.column_name,
                group_values=group_values,
                query_milliseconds=query_milliseconds)

    def get_value(self, results):
        value = None
        if self.metric_type == MonitorMetricType.MISSING_VALUES_PERCENTAGE:
            row_count = int(results[0])
            missing_count = int(results[1])
            value = None
            if row_count > 0:
                value = float(missing_count) * 100 / row_count

        elif self.metric_type in [MonitorMetricType.MISSING_VALUES_PERCENTAGE,
                                  MonitorMetricType.INVALID_VALUES_COUNT,
                                  MonitorMetricType.INVALID_VALUES_PERCENTAGE]:
            values_count = int(results[0])
            valid_count = int(results[1])
            invalid_count = values_count - valid_count
            value = None

            if self.metric_type == MonitorMetricType.VALID_VALUES_PERCENTAGE and values_count > 0:
                value = float(valid_count) * 100 / values_count
            elif self.metric_type == MonitorMetricType.INVALID_VALUES_COUNT:
                value = invalid_count
            elif self.metric_type == MonitorMetricType.INVALID_VALUES_PERCENTAGE and values_count > 0:
                value = float(invalid_count) * 100 / values_count

        elif self.metric_type == MonitorMetricType.UNIQUENESS_PERCENTAGE:
            valid_count = int(results[0])
            distinct_count = int(results[1])
            if valid_count > 1:
                value = (distinct_count - 1) * 100 / (valid_count - 1)
        else:
            value = results[0]
            if not (value is None or isinstance(value, int) or isinstance(value, float)):
                value = float(value)
        return value



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json
import logging
from typing import Optional, List

import requests

from sodasql.scan.scan_error import ScanError
from sodasql.scan.scan_yml import ScanYml
from sodasql.scan.scan_yml_column import ScanYmlColumn
from sodasql.__version__ import SODA_SQL_VERSION

logger = logging.getLogger(__name__)


class SodaServerClient:

    def __init__(self,
                 host: str,
                 port: Optional[str] = None,
                 protocol: Optional[str] = 'https',
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 api_key_id: Optional[str] = None,
                 api_key_secret: Optional[str] = None,
                 token: Optional[str] = None):
        self.host: str = host
        colon_port = f':{port}' if port else ''
        self.api_url: str = f'{protocol}://{self.host}{colon_port}/api'
        self.username: Optional[str] = username
        self.password: Optional[str] = password
        self.api_key_id: Optional[str] = api_key_id
        self.api_key_secret: Optional[str] = api_key_secret
        self.token: Optional[str] = token

    def scan_start(self, warehouse, scan_yml: ScanYml, scan_time,
                   origin: str = 'external'):
        soda_column_cfgs = {}
        if scan_yml.columns:
            for column_name in scan_yml.columns:
                scan_yml_column: ScanYmlColumn = scan_yml.columns[column_name]
                soda_column_cfg = {}
                if scan_yml_column.missing:
                    if scan_yml_column.missing.values:
                        soda_column_cfg['missingValues'] = scan_yml_column.missing.values
                    if scan_yml_column.missing.values:
                        soda_column_cfg['missingRegex'] = scan_yml_column.missing.regex
                    if scan_yml_column.missing.values:
                        soda_column_cfg['missingFormat'] = scan_yml_column.missing.format
                validity = scan_yml_column.validity
                if validity:
                    soda_column_cfg['validity'] = {
                        'namedFormat': validity.format,
                        'regexFormat': validity.regex,
                        'validValues': validity.values,
                        'minLength': validity.min_length,
                        'maxLength': validity.max_length,
                        'minValue': validity.min,
                        'maxValue': validity.max
                    }
                soda_column_cfgs[column_name] = soda_column_cfg

        return self.execute_command({
            'type': 'sodaSqlScanStart',
            'warehouseName': warehouse.name,
            'warehouseType': warehouse.dialect.type,
            'tableName': scan_yml.table_name,
            'scanTime': scan_time,
            'columns': soda_column_cfgs,
            'origin': origin
        })

    def scan_ended(self, scan_reference, errors: List[ScanError] = None):
        scan_end_command = {
            'type': 'sodaSqlScanEnd',
            'scanReference': scan_reference
        }
        if errors:
            logger.debug(f'Soda Cloud scan end with errors')
            scan_end_command['errors'] = [error.to_dict() for error in errors]
        else:
            logger.debug(f'Soda Cloud scan end ok')
        self.execute_command(scan_end_command)

    def scan_measurements(self, scan_reference: dict, measurement_jsons: List[dict]):
        logger.debug(f'Soda Cloud scan send measurements')
        return self.execute_command({
            'type': 'sodaSqlScanMeasurements',
            'scanReference': scan_reference,
            'measurements': measurement_jsons
        })

    def scan_test_results(self, scan_reference: dict, test_result_jsons: list):
        logger.debug(f'Soda Cloud scan send test results')
        return self.execute_command({
            'type': 'sodaSqlScanTestResults',
            'scanReference': scan_reference,
            'testResults': test_result_jsons
        })

    def scan_upload(self, scan_reference: str, file_path, temp_file, file_size_in_bytes: int):
        headers = {
            'Authorization': self.get_token(),
            'Content-Type': 'application/octet-stream',
            'Scan-Reference': scan_reference,
            'File-Path': file_path
        }

        if file_size_in_bytes == 0:
            # because of https://github.com/psf/requests/issues/4215 we can't send content size
            # when the size is 0 since requests blocks then on I/O indefinitely
            logger.warning("Empty file upload detected, not sending Content-Length header")
        else:
            headers['Content-Length'] = str(file_size_in_bytes)

        upload_response_json = self._upload_file(headers, temp_file)

        if 'fileId' not in upload_response_json:
            logger.error(f"No fileId received in response: {upload_response_json}")
        return upload_response_json['fileId']

    def _upload_file(self, headers, temp_file):
        upload_response = requests.post(
            f'{self.api_url}/scan/upload',
            headers=headers,
            data=temp_file)
        return upload_response.json()

    def scan_file(self,
                  scan_reference: str,
                  sample_type: str,
                  stored: int,
                  total: int,
                  source_columns: List[dict],
                  file_id: str,
                  column_name: Optional[str],
                  test_ids: List[str],
                  sql_metric_name: Optional[str] = None,
                  custom_metric_id: str = None):
        command_json = {
            'type': 'sodaSqlScanFile',
            'scanReference': scan_reference,
            'fileId': file_id,
            'sampleType': sample_type,
            'stored': stored,
            'total': total,
            'sourceColumns': source_columns
        }
        if column_name:
            command_json['columnName'] = column_name
        if test_ids:
            command_json['testSodaSqlIds'] = test_ids
        if sql_metric_name:
            command_json['sqlMetricName'] = sql_metric_name
        if custom_metric_id:
            command_json['customMetricId'] = custom_metric_id
        self.execute_command(command_json)

    def scan_monitor_measurements(self, scan_reference: dict, monitor_measurement_json: dict):
        return self.execute_command({
            'type': 'sodaSqlMonitorMeasurement',
            'scanReference': scan_reference,
            'monitorMeasurement': monitor_measurement_json
        })

    def historic_metrics(self, warehouse, table_name,  metrics):
        return self.execute_query({
            'type': 'sodaSqlHistoricMeasurements',
            'warehouseName': warehouse.name,
            'tableName': table_name,
            'metrics': metrics
        })

    def custom_metrics(self, scan_reference: dict):
        return self.execute_query({
            'type': 'sodaSqlCustomMetrics',
            'scanReference': scan_reference
        })

    def execute_command(self, command: dict):
        logger.debug(f'executing {command}')
        return self._execute_request('command', command, False)

    def execute_query(self, command: dict):
        return self._execute_request('query', command, False)

    def _execute_request(self, request_type: str, request_body: dict, is_retry: bool):
        # logger.debug(f'> /api/{request_type} {json.dumps(request_body, indent=2)}')
        request_body['token'] = self.get_token()
        request_body['sodaSqlVersion'] = SODA_SQL_VERSION
        response = requests.post(f'{self.api_url}/{request_type}', json=request_body)
        response_json = response.json()
        # logger.debug(f'< {response.status_code} {json.dumps(response_json, indent=2)}')
        if response.status_code == 401 and not is_retry:
            logger.debug(f'Authentication failed. Probably token expired. Reauthenticating...')
            self.token = None
            response_json = self._execute_request(request_type, request_body, True)
        else:
            assert response.status_code == 200, f'Request failed with status {response.status_code}: {json.dumps(response_json, indent=2)}'
        return response_json

    def get_token(self):
        if not self.token:
            login_command = {
                'type': 'login'
            }
            if self.api_key_id and self.api_key_secret:
                logger.debug('> /api/command (login with API key credentials)')
                login_command['apiKeyId'] = self.api_key_id
                login_command['apiKeySecret'] = self.api_key_secret
            elif self.username and self.password:
                logger.debug('> /api/command (login with username and password)')
                login_command['username'] = self.username
                login_command['password'] = self.password
            else:
                raise RuntimeError('No authentication in environment variables')

            login_response = requests.post(f'{self.api_url}/command', json=login_command)

            if login_response.status_code != 200:
                raise AssertionError(f'< {login_response.status_code} Login failed: {login_response.content}')
            login_response_json = login_response.json()
            self.token = login_response_json.get('token')
            assert self.token, 'No token in login response?!'
            logger.debug('< 200 (login ok, token received)')
        return self.token



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from dataclasses import dataclass
from typing import Optional, List
from deprecated import deprecated
from sodasql.common.json_helper import JsonHelper
from sodasql.scan.group_value import GroupValue


@dataclass
class Measurement:
    metric: str
    column_name: Optional[str] = None
    value: object = None
    group_values: Optional[List[GroupValue]] = None

    def __str__(self):
        column_str = f'({self.column_name})' if self.column_name else ''

        if self.group_values is not None:
            if len(self.group_values) == 0:
                return f'{self.metric}{column_str}: no groups'
            else:
                values_str = '\n  '.join([f'group{JsonHelper.to_jsonnable(group_value.group)} = {group_value.value}'
                                          for group_value in self.group_values])
                return f'{self.metric}{column_str}: \n  {values_str}'
        else:
            return f'{self.metric}{column_str} = {self.value}'

    def to_dict(self) -> dict:
        dictionary = {
            'metric': self.metric,
        }

        if self.group_values is None:
            dictionary['value'] = JsonHelper.to_jsonnable(self.value)
        else:
            dictionary['groupValues'] = [group_value.to_dict() for group_value in self.group_values]

        if self.column_name is not None:
            dictionary['columnName'] = self.column_name

        return dictionary

    @deprecated(version='2.1.0b19', reason='This function is deprecated, please use to_dict')
    def to_json(self):
        return self.to_dict()



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from dataclasses import dataclass
from typing import List

from sodasql.scan.validity import Validity
from sodasql.scan.warehouse import Warehouse
from deprecated import deprecated


@dataclass
class ColumnAnalysisResult:
    column_name: str
    source_type: str
    is_text: bool = False
    validity_format: str = None
    values_count: int = None
    valid_count: int = None

    def to_dict(self):
        return {
            'columnName': self.column_name,
            'sourceType': self.source_type,
            'validityFormat': self.validity_format
        }

    @deprecated(version='2.1.0b19', reason='This function is deprecated, please use to_dict')
    def to_json(self):
        return self.to_dict()


class DatasetAnalyzer:

    def analyze(self, warehouse: Warehouse, table_name: str):
        dialect = warehouse.dialect
        qualified_table_name = dialect.qualify_table_name(table_name)

        analyze_results: List[ColumnAnalysisResult] = []

        sql = dialect.sql_columns_metadata_query(table_name)
        column_tuple_list = dialect.sql_columns_metadata(table_name)
        column_tuples = warehouse.sql_fetchall(sql) if len(
            column_tuple_list) == 0 else column_tuple_list
        for column_tuple in column_tuples:
            column_name = column_tuple[0]
            source_type = column_tuple[1]

            column_analysis_result = ColumnAnalysisResult(
                column_name=column_name, source_type=source_type)
            analyze_results.append(column_analysis_result)

            qualified_column_name = dialect.qualify_column_name(column_name)
            select_with_limit_query = dialect.sql_select_with_limit(qualified_table_name, 1000)

            if dialect.is_text(source_type):
                column_analysis_result.is_text = True

                validity_format_count_fields = []
                validity_counts = []
                for validity_format in Validity.FORMATS:
                    format_regex = Validity.FORMATS[validity_format]
                    validity_counts.append({'format': validity_format})
                    qualified_regex = dialect.qualify_regex(format_regex)
                    regexp_like = dialect.sql_expr_regexp_like(
                        qualified_column_name, qualified_regex)
                    count_field = f'COUNT(CASE WHEN {regexp_like} THEN 1 END) AS {validity_format}'
                    validity_format_count_fields.append(count_field)

                row = warehouse.sql_fetchone(
                    f'SELECT \n  ' +
                    (',\n  '.join(validity_format_count_fields)) + ',\n'
                                                                   f'  COUNT({qualified_column_name}) \n'
                                                                   f'FROM ({select_with_limit_query}) T'
                )

                values_count = row[len(validity_counts)]
                column_analysis_result.values_count = values_count

                if values_count > 0:
                    for i in range(len(validity_counts)):
                        validity_count = validity_counts[i]
                        validity_count['count'] = row[i]

                    sorted_validity_counts = sorted(
                        validity_counts, key=lambda c: c['count'], reverse=True)
                    most_frequent_validity_format = sorted_validity_counts[0]
                    valid_count = most_frequent_validity_format['count']
                    column_analysis_result.valid_count = valid_count

                    if valid_count > (values_count / 2):
                        column_analysis_result.validity_format = most_frequent_validity_format[
                            'format']

        return analyze_results



#   Copyright 2020 Soda
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

ERROR_CODE_GENERIC = 'generic_error'
ERROR_CODE_CONNECTION_FAILED = 'connection_failed'
ERROR_CODE_AUTHENTICATION_FAILED = 'authentication_failed'
ERROR_CODE_TEST_FAILED = 'test_failed'


class SodaSqlError(Exception):

    def __init__(self, msg, original_exception):
        super(SodaSqlError, self).__init__(f"{msg}: {str(original_exception)}")
        self.error_code = ERROR_CODE_GENERIC
        self.original_exception = original_exception


class WarehouseAuthenticationError(SodaSqlError):

    def __init__(self, warehouse_type, original_exception):
        super(WarehouseAuthenticationError, self).__init__(
            f"Soda-sql encountered a problem while trying to authenticate to {warehouse_type}",
            original_exception)
        self.error_code = ERROR_CODE_AUTHENTICATION_FAILED
        self.warehouse_type = warehouse_type


class WarehouseConnectionError(SodaSqlError):

    def __init__(self, warehouse_type, original_exception):
        super(WarehouseConnectionError, self).__init__(
            f"Soda-sql encountered a problem while trying to connect to {warehouse_type}",
            original_exception)
        self.error_code = ERROR_CODE_CONNECTION_FAILED
        self.warehouse_type = warehouse_type


class TestFailureError(SodaSqlError):

    def __init__(self, original_exception, errors_count):
        if errors_count > 1:
            msg = f"{errors_count} soda-sql tests failed with errors"
        else:
            msg = "Soda-sql test failed with error"
        super(TestFailureError, self).__init__(msg,
                                               original_exception)
        self.error_code = ERROR_CODE_TEST_FAILED
        self.errors_number = errors_count

class InvalidWarehouseYaml(Exception):

    def __init__(self, exception_detail):
        super(InvalidWarehouseYaml, self).__init__(
            f"Invalid Warehouse YAML. {exception_detail}")


#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from dataclasses import dataclass
from typing import Optional, List
from jinja2 import Template

logger = logging.getLogger(__name__)

@dataclass
class Test:
    id: str
    title: str
    expression: str
    metrics: List[str]
    column: Optional[str]
    expression_delimiters = ['<=', '>=', '<', '>', '==']

    def evaluate(self, test_variables: dict, group_values: Optional[dict] = None,
                 template_variables: Optional[dict] = None):
        from sodasql.scan.test_result import TestResult

        try:
            values = {key: test_variables[key] for key in test_variables if key in self.metrics}
            if template_variables is not None:
                self.expression = Template(self.expression).render(template_variables)
            if 'None' not in self.expression and any(v is None for v in values.values()):
                logger.warning(f'Skipping test {self.expression} since corresponding metrics are None ({values}) ')
                return TestResult(test=self, skipped=True, passed=True, values=values, group_values=group_values)
            else:
                passed = bool(eval(self.expression, test_variables))

                # Evaluate more complex expressions and save result of the expression.
                for delimiter in self.expression_delimiters:
                    if delimiter in self.expression:
                        left, _, _ = self.expression.partition(delimiter)
                        # Make sure the expression result is the first key in the resulting dict.
                        expression_result = {'expression_result': eval(left, test_variables)}
                        expression_result.update(values)
                        values = expression_result

                        break

                test_result = TestResult(test=self, passed=passed, skipped=False, values=values,
                                         group_values=group_values)
                logger.debug(str(test_result))
                return test_result
        except Exception as e:
            logger.error(f'Test error for "{self.expression}": {e}')
            return TestResult(test=self, passed=False, skipped=False, error=e, group_values=group_values)



from functools import wraps
from os import name
import textwrap
from typing import Any, Dict, List, Optional

from opentelemetry import trace
from opentelemetry.trace.span import Span

import ast
import inspect

from opentelemetry.trace.status import Status, StatusCode
from opentelemetry.trace.propagation.tracecontext import \
    TraceContextTextMapPropagator

from sodasql.telemetry.soda_telemetry import SodaTelemetry
trace_context_propagator = TraceContextTextMapPropagator()
trace_context_carrier = {}

soda_telemetry = SodaTelemetry.get_instance()
tracer = trace.get_tracer_provider().get_tracer(__name__)


def get_decorators(function):
    """
    Fancy introspection - very WIP
    """
    decorators = {}

    def visit_FunctionDef(node):
        decorators[node.name] = {}
        for n in node.decorator_list:
            print(ast.dump(n))
            name = ''
            if isinstance(n, ast.Call):
                group = n.func.value.id
                name = n.func.attr if isinstance(n.func, ast.Attribute) else n.func.id

                for a in n.args:
                    print(ast.dump(a))
            else:
                group = n.attr if isinstance(n, ast.Attribute) else n.id
                name = None

            if group not in decorators[node.name]:
                decorators[node.name][group] = []

            if name:
                decorators[node.name][group].append({name})

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_FunctionDef
    node_iter.visit(ast.parse(textwrap.dedent(inspect.getsource(function))))
    return decorators



def soda_trace(fn: callable):
    def _before_exec(span: Span, fn: callable):
        span.set_attribute("user_cookie_id", soda_telemetry.user_cookie_id)

    def _after_exec(span: Span, error: Optional[BaseException] = None):
        if str(error) == "0":
            pass
            # This is an OK cli exit state
            span.set_status(Status(StatusCode.OK))
        else:
            span.set_status(Status(StatusCode.ERROR))
            # TODO: this will now produce an "error" attribute with value "1" for CLI as all Exceptions are caught and re-raised as system.exit(1). Revisit error handling in CLI for more useful info in traces.
            span.add_event("error", {"error": str(error)})


    @wraps(fn)
    def wrapper(*original_args, **original_kwargs):
        ctx = trace_context_propagator.extract(carrier=trace_context_carrier)
        with tracer.start_as_current_span(f"{fn.__module__}.{fn.__name__}", context=ctx) as span:
            trace_context_propagator.inject(carrier=trace_context_carrier)
            result = None
            try:
                _before_exec(span, fn)
                result = fn(*original_args, **original_kwargs)
                span.set_status(Status(StatusCode.OK))
                _after_exec(span)
            except BaseException as e:
                # Catch base exception to deal with exit codes as well.
                _after_exec(span, e)
        return result
    return wrapper

def span_setup_function_args(args: Dict):
    for prefix, values in args.items():
        for key, value in values.items():
            soda_telemetry.set_attribute(f'{prefix}_{key}', value or "")



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


class FailedRowsProcessor:
    def process(self, context: dict):
        """
        Override this class in your program to process the failed rows
        :param context: dict with following keys:

            column_name : Name of the Column
            sample_name : Generated name of the sample
            sample_columns: Columns of the selected samples
            sample_rows: List of rows
            sample_description: Auto-generated description of the samples
            total_row_count: total count of the failed rows

        :return: dict with  message and count keys which will be sent to soda cloud instead of failed rows
        e.g. {'message': 'Failed Rows are saved in S3 bucket s3:///test_failed_rows'
             'count': 23}
        """
        pass



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from sodasql.scan.dialect import Dialect
from sodasql.scan.parser import Parser
from sodasql.scan.scan import Scan

from sodasql.soda_server_client.monitor_metric import MonitorMetric

KEY_ID = 'id'
KEY_TYPE = 'type'
KEY_COLUMN_NAME = 'columnName'
KEY_GROUP_BY_COLUMN_NAMES = 'groupByColumnNames'
KEY_FILTER = 'filter'


class MonitorMetricParser(Parser):

    def __init__(self, monitor_metric_dict: dict, scan: Scan):
        super().__init__('Monitor SQL metric')
        self.dialect: Dialect = scan.dialect
        self._push_context(monitor_metric_dict)
        try:
            metric_type = self.get_str_required(KEY_TYPE)
            metric_id = self.get_str_required(KEY_ID)
            column_name = self.get_str_optional(KEY_COLUMN_NAME)
            group_by_column_names = self.get_list_optional(KEY_GROUP_BY_COLUMN_NAMES)

            self.monitor_metric = MonitorMetric(
                scan=scan,
                metric_id=metric_id,
                metric_type=metric_type,
                column_name=column_name,
                group_by_column_names=group_by_column_names
            )

            filter_expression_dict = self.get_dict_optional(KEY_FILTER)

            qualified_group_column_names = []
            if group_by_column_names:
                qualified_group_column_names = [self.dialect.qualify_column_name(group_field)
                                                for group_field in group_by_column_names]
            filter_condition = self.dialect.sql_expression(expression_dict=filter_expression_dict,
                                                           scan_time=scan.time)

            self.monitor_metric.build_sql(
                qualified_group_column_names,
                filter_condition,
                scan.qualified_table_name)

        finally:
            self._pop_context()



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import os
import json
import tempfile
from datetime import datetime
from math import floor, ceil
from typing import List, Optional

from jinja2 import Template
from sodasql.scan.historic_metric_yml import HistoricMetricYml

from sodasql.scan.column_metadata import ColumnMetadata
from sodasql.scan.group_value import GroupValue
from sodasql.scan.measurement import Measurement
from sodasql.scan.metric import Metric
from sodasql.scan.scan_column import ScanColumn
from sodasql.scan.scan_error import SodaCloudScanError, ScanError, TestExecutionScanError
from sodasql.scan.scan_result import ScanResult
from sodasql.scan.scan_yml import ScanYml
from sodasql.scan.sql_metric_yml import SqlMetricYml
from sodasql.scan.test import Test
from sodasql.scan.test_result import TestResult
from sodasql.scan.warehouse import Warehouse
from sodasql.soda_server_client.soda_server_client import SodaServerClient
from sodasql.scan.failed_rows_processor import FailedRowsProcessor

logger = logging.getLogger(__name__)


class Scan:

    def __init__(self,
                 warehouse: Warehouse,
                 scan_yml: ScanYml = None,
                 soda_server_client: SodaServerClient = None,
                 variables: dict = None,
                 time: str = None,
                 scan_results_file: Optional[str] = None,
                 failed_rows_processor: Optional[FailedRowsProcessor] = None):
        self.warehouse = warehouse
        self.scan_yml = scan_yml
        self.soda_server_client = soda_server_client
        self.variables = variables
        self.time = time
        self.scan_results_file = scan_results_file
        self.failed_rows_processor = failed_rows_processor

        self.scan_result = ScanResult()
        self.dialect = warehouse.dialect
        self.qualified_table_name = self.dialect.qualify_table_name(scan_yml.table_name)
        self.scan_reference = None
        self.disable_sample_collection = False
        self.column_metadatas: List[ColumnMetadata] = []
        self.column_names: List[str] = []
        # maps column names (lower case) to ScanColumn's
        self.scan_columns: dict = {}
        self.close_warehouse = True
        self.send_scan_end = True
        self.start_time = None
        self.queries_executed = 0
        self.sampler = None

        self.table_sample_clause = \
            f'\nTABLESAMPLE {scan_yml.sample_method}({scan_yml.sample_percentage})' \
                if scan_yml.sample_percentage \
                else ''

        self.filter_sql = None
        if scan_yml.filter_template:
            if not variables:
                raise RuntimeError(f'No variables provided while filter "{str(scan_yml.filter)}" specified')
            self.filter_sql = scan_yml.filter_template.render(variables)

    def execute(self) -> ScanResult:
        self.start_time = datetime.now()
        if self.soda_server_client:
            logger.debug(f'Soda cloud: {self.soda_server_client.host}')
            self._ensure_scan_reference()

        from sodasql.scan.sampler import Sampler
        self.sampler = Sampler(self)

        try:
            if self.scan_yml:
                # Soda Server and the code below require that the schema measurements is the first measurement
                self._query_columns_metadata()
                self._query_aggregations()
                self._query_group_by_value()
                self._query_histograms()

            self._query_sql_metrics_and_run_tests()
            self._process_cloud_custom_metrics()
            self._run_table_tests()
            self._run_column_tests()
            if not self.disable_sample_collection:
                self._process_samples()

            logger.debug(f'Executed {self.queries_executed} queries in {(datetime.now() - self.start_time)}')

        except Exception as e:
            logger.exception('Exception during scan')
            self.scan_result.add_error(ScanError('Exception during scan', e))

        finally:
            if self.soda_server_client and self.send_scan_end:
                try:
                    self.soda_server_client.scan_ended(self.scan_reference, self.scan_result.errors)
                except Exception as e:
                    logger.error(f'Soda Cloud error: Could not end scan: {e}')
                    self.scan_result.add_error(SodaCloudScanError('Could not end scan', e))

            if self.close_warehouse:
                self.warehouse.close()

        if self.scan_results_file is not None:
            logger.info(f'Saving scan results to {self.scan_results_file}')
            with open(self.scan_results_file, 'w') as f:
                json.dump(self.scan_result.to_dict(), f)

        return self.scan_result

    def _process_cloud_custom_metrics(self):
        if self.soda_server_client:
            from sodasql.soda_server_client.monitor_metric_parser import MonitorMetricParser
            from sodasql.soda_server_client.monitor_metric import MonitorMetricType
            custom_metrics = []
            try:
                logger.debug(f'Fetching custom metrics with scanReference: {self.scan_reference}')
                custom_metrics = self.soda_server_client.custom_metrics(self.scan_reference)
            except Exception as e:
                logger.error(f'Soda cloud error: Could not fetch custom_metrics: {e}')
                self.scan_result.add_error(SodaCloudScanError('Could not fetch custom metrics', e))
            if custom_metrics:
                for monitor_metric_dict in custom_metrics:
                    monitor_metric_parser = MonitorMetricParser(monitor_metric_dict, self)
                    if not monitor_metric_parser.has_warnings_or_errors():
                        monitor_metric = monitor_metric_parser.monitor_metric
                        monitor_measurement = monitor_metric.execute()
                        self.scan_result.measurements.append(monitor_measurement)
                        monitor_measurement_json = monitor_measurement.to_dict()
                        self.soda_server_client.scan_monitor_measurements(self.scan_reference,
                                                                          monitor_measurement_json)
                        if self.disable_sample_collection:
                            logger.info(
                                'failed rows are not sent to Soda Cloud because sample collection is disabled')
                        else:
                            if monitor_metric.metric_type in [
                                MonitorMetricType.MISSING_VALUES_COUNT,
                                MonitorMetricType.MISSING_VALUES_PERCENTAGE,
                                MonitorMetricType.INVALID_VALUES_COUNT,
                                MonitorMetricType.INVALID_VALUES_PERCENTAGE,
                                MonitorMetricType.DISTINCT_VALUES_COUNT]:
                                logger.debug(
                                    f'Sending failed rows for {monitor_metric.metric_type} to Soda Cloud')
                                self._send_failed_rows_custom_metric(metric_name="custom_metric",
                                                                     sql=monitor_metric.sql,
                                                                     column_name=monitor_metric.column_name,
                                                                     metric_id=monitor_metric.metric_id)
                            else:
                                logger.info(
                                    f'{monitor_metric.metric_type} is not "negative values metric", failed rows are not sent to Soda Cloud')

    def _query_columns_metadata(self):
        sql = self.warehouse.dialect.sql_columns_metadata_query(self.scan_yml.table_name)
        column_tuples = self.warehouse.sql_fetchall(sql) if sql != '' else self.warehouse.dialect.sql_columns_metadata(
            self.scan_yml.table_name)
        self.queries_executed += 1

        self.column_metadatas = []
        for column_tuple in column_tuples:
            name = column_tuple[0]
            data_type = column_tuple[1]
            nullable = 'YES' == column_tuple[2].upper()
            self.column_metadatas.append(ColumnMetadata(name=name, data_type=data_type, nullable=nullable))

        self.scan_columns: dict = {}
        self.column_names: List[str] = [column_metadata.name for column_metadata in self.column_metadatas]
        for column_metadata in self.column_metadatas:
            if (self.scan_yml.excluded_columns is not None) and (
                column_metadata.name in self.scan_yml.excluded_columns):
                logger.info(f'  column {column_metadata.name} is excluded based on configuration!')

            else:
                scan_column = ScanColumn(self, column_metadata)
                if scan_column.is_supported:
                    if scan_column.is_numeric:
                        column_metadata.logical_type = 'number'
                    elif scan_column.is_temporal:
                        column_metadata.logical_type = 'time'
                    # WARN: Order of conditions is important here, because the is_text is based on database data type,
                    # and is_numeric, is_temporal are calculated based on validity format. So both is_text and
                    # (is_temporal|is_numeric) can be true.
                    elif scan_column.is_text:
                        column_metadata.logical_type = 'text'

                    logger.debug(f'  {scan_column.column_name} ({scan_column.column.data_type}) '
                                 f'{"" if scan_column.column.nullable else "not null"}')
                    self.scan_columns[column_metadata.name.lower()] = scan_column
                else:
                    logger.info(
                        f'  {scan_column.column_name} ({scan_column.column.data_type}) -> unsupported, skipped!')
        logger.debug(str(len(self.column_metadatas)) + ' columns:')

        # Compare the column names in yml with valid column names from the table
        invalid_column_names = set(self.scan_yml.columns.keys()) - set(self.scan_columns.keys())
        if invalid_column_names:
            logger.warning(f'Unknown or unsupported columns found in YML: {", ".join(invalid_column_names)} \n'
                           f'Scan will continue for known/supported columns.')

        schema_measurement_value = [column_metadata.to_dict() for column_metadata in self.column_metadatas]
        schema_measurement = Measurement(Metric.SCHEMA, value=schema_measurement_value)
        self._log_measurement(schema_measurement)
        self._flush_measurements([schema_measurement])

    def _query_aggregations(self):
        # This measurements list is used to match measurements with the query field order.
        # After query execution, the value of the measurements will be extracted from the query result and
        # the measurements will be added with self.add_query(measurement)
        measurements: List[Measurement] = []

        fields: List[str] = []

        dialect = self.warehouse.dialect

        if self.scan_yml.is_metric_enabled(Metric.ROW_COUNT):
            fields.append(dialect.sql_expr_count_all())
            measurements.append(Measurement(Metric.ROW_COUNT))

        # maps db column names (lower) to missing and invalid metric indices in the measurements
        # eg { 'colname': {'missing': 2, 'invalid': 3}, ...}
        column_metric_indices = {}

        try:
            for column_name_lower, scan_column in self.scan_columns.items():
                metric_indices = {}
                column_metric_indices[column_name_lower] = metric_indices
                column_name = scan_column.column_name

                if scan_column.is_missing_enabled:
                    metric_indices['non_missing'] = len(measurements)
                    if scan_column.non_missing_condition:
                        fields.append(dialect.sql_expr_count_conditional(scan_column.non_missing_condition))
                    else:
                        fields.append(dialect.sql_expr_count(scan_column.qualified_column_name))
                    measurements.append(Measurement(Metric.VALUES_COUNT, column_name))

                if scan_column.is_valid_enabled:
                    metric_indices['valid'] = len(measurements)
                    if scan_column.non_missing_and_valid_condition:
                        fields.append(dialect.sql_expr_count_conditional(scan_column.non_missing_and_valid_condition))
                    else:
                        fields.append(dialect.sql_expr_count(scan_column.qualified_column_name))
                    measurements.append(Measurement(Metric.VALID_COUNT, column_name))

                if scan_column.is_text:
                    length_expr = dialect.sql_expr_conditional(
                        scan_column.non_missing_and_valid_condition,
                        dialect.sql_expr_length(scan_column.qualified_column_name)) \
                        if scan_column.non_missing_and_valid_condition \
                        else dialect.sql_expr_length(scan_column.qualified_column_name)

                    if self.scan_yml.is_metric_enabled(Metric.AVG_LENGTH, column_name):
                        fields.append(dialect.sql_expr_avg(length_expr))
                        measurements.append(Measurement(Metric.AVG_LENGTH, column_name))

                    if self.scan_yml.is_metric_enabled(Metric.MIN_LENGTH, column_name):
                        fields.append(dialect.sql_expr_min(length_expr))
                        measurements.append(Measurement(Metric.MIN_LENGTH, column_name))

                    if self.scan_yml.is_metric_enabled(Metric.MAX_LENGTH, column_name):
                        fields.append(dialect.sql_expr_max(length_expr))
                        measurements.append(Measurement(Metric.MAX_LENGTH, column_name))

                if scan_column.is_numeric:
                    if scan_column.is_metric_enabled(Metric.MIN):
                        fields.append(dialect.sql_expr_min(scan_column.numeric_expr))
                        measurements.append(Measurement(Metric.MIN, column_name))

                    if scan_column.is_metric_enabled(Metric.MAX):
                        fields.append(dialect.sql_expr_max(scan_column.numeric_expr))
                        measurements.append(Measurement(Metric.MAX, column_name))

                    if scan_column.is_metric_enabled(Metric.AVG):
                        fields.append(dialect.sql_expr_avg(scan_column.numeric_expr))
                        measurements.append(Measurement(Metric.AVG, column_name))

                    if scan_column.is_metric_enabled(Metric.SUM):
                        fields.append(dialect.sql_expr_sum(scan_column.numeric_expr))
                        measurements.append(Measurement(Metric.SUM, column_name))

                    if scan_column.is_metric_enabled(Metric.VARIANCE):
                        fields.append(dialect.sql_expr_variance(scan_column.numeric_expr))
                        measurements.append(Measurement(Metric.VARIANCE, column_name))

                    if scan_column.is_metric_enabled(Metric.STDDEV):
                        fields.append(dialect.sql_expr_stddev(scan_column.numeric_expr))
                        measurements.append(Measurement(Metric.STDDEV, column_name))

            if len(fields) > 0:
                sql = 'SELECT \n  ' + ',\n  '.join(fields) + ' \n' \
                                                             'FROM ' + self.qualified_table_name
                if self.table_sample_clause:
                    sql += f'\n{self.table_sample_clause}'
                if self.filter_sql:
                    sql += f'\nWHERE {self.filter_sql}'

                query_result_tuple = self.warehouse.sql_fetchone(sql)
                self.queries_executed += 1

                for i in range(0, len(measurements)):
                    measurement = measurements[i]
                    measurement.value = query_result_tuple[i]
                    self._log_measurement(measurement)

                # Calculating derived measurements
                row_count_measurement = next((m for m in measurements if m.metric == Metric.ROW_COUNT), None)
                if row_count_measurement:
                    row_count = row_count_measurement.value
                    for column_name_lower, scan_column in self.scan_columns.items():
                        column_name = scan_column.column_name
                        metric_indices = column_metric_indices[column_name_lower]
                        non_missing_index = metric_indices.get('non_missing')
                        if non_missing_index is not None:
                            values_count = measurements[non_missing_index].value
                            missing_count = row_count - values_count
                            missing_percentage = missing_count * 100 / row_count if row_count > 0 else None
                            values_percentage = values_count * 100 / row_count if row_count > 0 else None

                            self._log_and_append_derived_measurements(measurements, [
                                Measurement(Metric.MISSING_PERCENTAGE, column_name, missing_percentage),
                                Measurement(Metric.MISSING_COUNT, column_name, missing_count),
                                Measurement(Metric.VALUES_PERCENTAGE, column_name, values_percentage)
                            ])

                            valid_index = metric_indices.get('valid')
                            if valid_index is not None:
                                valid_count = measurements[valid_index].value
                                invalid_count = row_count - missing_count - valid_count
                                invalid_percentage = invalid_count * 100 / row_count if row_count > 0 else None
                                valid_percentage = valid_count * 100 / row_count if row_count > 0 else None
                                self._log_and_append_derived_measurements(measurements, [
                                    Measurement(Metric.INVALID_PERCENTAGE, column_name, invalid_percentage),
                                    Measurement(Metric.INVALID_COUNT, column_name, invalid_count),
                                    Measurement(Metric.VALID_PERCENTAGE, column_name, valid_percentage)
                                ])

            self._flush_measurements(measurements)
        except Exception as e:
            logger.debug(f'Exception during aggregation query', exc_info=e)
            self.scan_result.add_error(ScanError(f'Exception during aggregation query', e))

    def _query_group_by_value(self):
        for column_name_lower, scan_column in self.scan_columns.items():
            try:
                measurements = []

                column_name = scan_column.column_name

                if scan_column.is_any_metric_enabled(
                    [Metric.DISTINCT, Metric.UNIQUENESS, Metric.UNIQUE_COUNT,
                     Metric.MINS, Metric.MAXS, Metric.FREQUENT_VALUES, Metric.DUPLICATE_COUNT]):

                    group_by_cte = scan_column.get_group_by_cte()
                    numeric_value_expr = scan_column.get_group_by_cte_numeric_value_expression()
                    order_by_value_expr = scan_column.get_order_by_cte_value_expression(numeric_value_expr)

                    if self.scan_yml.is_any_metric_enabled(
                        [Metric.DISTINCT, Metric.UNIQUENESS, Metric.UNIQUE_COUNT, Metric.DUPLICATE_COUNT],
                        column_name):

                        sql = (f'{group_by_cte} \n'
                               f'SELECT COUNT(*), \n'
                               f'       COUNT(CASE WHEN frequency = 1 THEN 1 END), \n'
                               f'       SUM(frequency) \n'
                               f'FROM group_by_value')

                        query_result_tuple = self.warehouse.sql_fetchone(sql)
                        self.queries_executed += 1

                        distinct_count = query_result_tuple[0]
                        unique_count = query_result_tuple[1]
                        valid_count = query_result_tuple[2] if query_result_tuple[2] else 0
                        duplicate_count = distinct_count - unique_count

                        self._log_and_append_query_measurement(
                            measurements, Measurement(Metric.DISTINCT, column_name, distinct_count))
                        self._log_and_append_query_measurement(
                            measurements, Measurement(Metric.UNIQUE_COUNT, column_name, unique_count))

                        derived_measurements = [Measurement(Metric.DUPLICATE_COUNT, column_name, duplicate_count)]
                        if valid_count > 1:
                            uniqueness = (distinct_count - 1) * 100 / (valid_count - 1)
                            derived_measurements.append(Measurement(Metric.UNIQUENESS, column_name, uniqueness))
                        self._log_and_append_derived_measurements(measurements, derived_measurements)

                    if scan_column.is_metric_enabled(Metric.MINS) and order_by_value_expr:
                        sql = (f'{group_by_cte} \n'
                               f'SELECT value \n'
                               f'FROM group_by_value \n'
                               f'ORDER BY {order_by_value_expr} ASC \n'
                               f'{self.dialect.sql_expr_limit(scan_column.mins_maxs_limit)}\n')

                        rows = self.warehouse.sql_fetchall(sql)
                        self.queries_executed += 1

                        mins = [row[0] for row in rows]
                        self._log_and_append_query_measurement(measurements,
                                                               Measurement(Metric.MINS, column_name, mins))

                    if self.scan_yml.is_metric_enabled(Metric.MAXS, column_name) and order_by_value_expr:
                        sql = (f'{group_by_cte} \n'
                               f'SELECT value \n'
                               f'FROM group_by_value \n'
                               f'ORDER BY {order_by_value_expr} DESC \n'
                               f'{self.dialect.sql_expr_limit(scan_column.mins_maxs_limit)}\n')

                        rows = self.warehouse.sql_fetchall(sql)
                        self.queries_executed += 1

                        maxs = [row[0] for row in rows]
                        self._log_and_append_query_measurement(measurements,
                                                               Measurement(Metric.MAXS, column_name, maxs))

                    if self.scan_yml.is_metric_enabled(Metric.FREQUENT_VALUES, column_name):
                        frequent_values_limit = self.scan_yml.get_frequent_values_limit(column_name)
                        sql = (f'{group_by_cte} \n'
                               f'SELECT value, frequency \n'
                               f'FROM group_by_value \n'
                               f'ORDER BY frequency DESC \n'
                               f'{self.dialect.sql_expr_limit(scan_column.mins_maxs_limit)}\n')

                        rows = self.warehouse.sql_fetchall(sql)
                        self.queries_executed += 1
                        frequent_values = [{'value': row[0], 'frequency': row[1]} for row in rows]
                        self._log_and_append_query_measurement(
                            measurements, Measurement(Metric.FREQUENT_VALUES, column_name, frequent_values))

                self._flush_measurements(measurements)
            except Exception as e:
                self.scan_result.add_error(ScanError(f'Exception during column group by value queries', e))

    def _query_histograms(self):

        for column_name_lower, scan_column in self.scan_columns.items():
            column_name = scan_column.column_name
            try:
                if scan_column.is_metric_enabled(Metric.HISTOGRAM) and scan_column.numeric_expr:
                    measurements = []
                    buckets: int = scan_column.get_histogram_buckets()

                    min_value = scan_column.get_metric_value(Metric.MIN)
                    max_value = scan_column.get_metric_value(Metric.MAX)

                    if scan_column.is_numeric and min_value and max_value and min_value < max_value:
                        # Build the histogram query
                        min_value = floor(min_value * 1000) / 1000
                        max_value = ceil(max_value * 1000) / 1000
                        bucket_width = (max_value - min_value) / buckets

                        boundary = min_value
                        boundaries = [min_value]
                        for i in range(0, buckets):
                            boundary += bucket_width
                            boundaries.append(round(boundary, 3))

                        group_by_cte = scan_column.get_group_by_cte()
                        numeric_value_expr = scan_column.get_group_by_cte_numeric_value_expression()

                        field_clauses = []
                        for i in range(0, buckets):
                            lower_bound = '' if i == 0 else f'{boundaries[i]} <= {numeric_value_expr}'
                            upper_bound = '' if i == buckets - 1 else f'{numeric_value_expr} < {boundaries[i + 1]}'
                            optional_and = '' if lower_bound == '' or upper_bound == '' else ' and '
                            field_clauses.append(
                                f'SUM(CASE WHEN {lower_bound}{optional_and}{upper_bound} THEN frequency END)')

                        fields = ',\n  '.join(field_clauses)

                        sql = (f'{group_by_cte} \n'
                               f'SELECT \n'
                               f'  {fields} \n'
                               f'FROM group_by_value')

                        row = self.warehouse.sql_fetchone(sql)
                        self.queries_executed += 1

                        # Process the histogram query
                        frequencies = []
                        for i in range(0, buckets):
                            frequency = row[i]
                            frequencies.append(0 if not frequency else int(frequency))
                        histogram = {
                            'boundaries': boundaries,
                            'frequencies': frequencies
                        }

                        self._log_and_append_query_measurement(measurements,
                                                               Measurement(Metric.HISTOGRAM, column_name, histogram))
                        self._flush_measurements(measurements)
            except Exception as e:
                self.scan_result.add_error(ScanError(f'Exception during histogram query for {column_name}', e))

    def _query_sql_metrics_and_run_tests(self):
        self._query_sql_metrics_and_run_tests_base(self.scan_yml.sql_metric_ymls)
        for column_name_lower, scan_column in self.scan_columns.items():
            if scan_column and scan_column.scan_yml_column:
                self._query_sql_metrics_and_run_tests_base(scan_column.scan_yml_column.sql_metric_ymls, scan_column)

    def _query_sql_metrics_and_run_tests_base(self,
                                              sql_metric_ymls: Optional[List[SqlMetricYml]],
                                              scan_column: Optional[ScanColumn] = None):
        if sql_metric_ymls:
            for sql_metric in sql_metric_ymls:
                try:
                    resolved_sql = self.resolve_sql_metric_sql(sql_metric)

                    if sql_metric.type == 'numeric':
                        self._run_sql_metric_default_and_run_tests(sql_metric, resolved_sql, scan_column)
                    elif sql_metric.type == 'numeric_groups':
                        self._run_sql_metric_with_groups_and_run_tests(sql_metric, resolved_sql, scan_column)
                    elif sql_metric.type == 'failed_rows' and not self.disable_sample_collection:
                        self._run_sql_metric_failed_rows(sql_metric, resolved_sql, scan_column)
                except Exception as e:
                    msg = f"Couldn't run sql metric {sql_metric}: {e}"
                    logger.exception(msg)
                    self.scan_result.add_error(ScanError(f'Could not run sql metric {sql_metric}', e))

    def resolve_sql_metric_sql(self, sql_metric):
        if self.variables:
            sql_variables = self.variables.copy() if self.variables else {}
            template = Template(sql_metric.sql)
            resolved_sql = template.render(sql_variables)
        else:
            resolved_sql = sql_metric.sql
        return resolved_sql

    def _run_sql_metric_default_and_run_tests(self,
                                              sql_metric: SqlMetricYml,
                                              resolved_sql: str,
                                              scan_column: Optional[ScanColumn] = None):
        try:
            row_tuple, description = self.warehouse.sql_fetchone_description(resolved_sql)
            self.queries_executed += 1

            test_variables = self._get_test_variables(scan_column)
            column_name = scan_column.column_name if scan_column is not None else None

            measurements = []
            for i in range(len(row_tuple)):
                metric_name = sql_metric.metric_names[i] if sql_metric.metric_names is not None else description[i][0]
                metric_value = row_tuple[i]
                logger.debug(f'SQL metric {sql_metric.title} {metric_name} -> {metric_value}')
                measurement = Measurement(metric=metric_name, value=metric_value, column_name=column_name)
                test_variables[metric_name] = metric_value
                self._log_and_append_query_measurement(measurements, measurement)

            self._flush_measurements(measurements)

            sql_metric_test_results = self._execute_tests(sql_metric.tests, test_variables)
            self._flush_test_results(sql_metric_test_results)
        except Exception as e:
            self.scan_result.add_error(ScanError(f'Exception during sql metric query {resolved_sql}', e))

    def _run_sql_metric_with_groups_and_run_tests(self,
                                                  sql_metric: SqlMetricYml,
                                                  resolved_sql: str,
                                                  scan_column: Optional[ScanColumn]):
        try:
            measurements = []
            test_results = []
            group_fields_lower = set(group_field.lower() for group_field in sql_metric.group_fields)

            rows, description = self.warehouse.sql_fetchall_description(resolved_sql)
            self.queries_executed += 1

            group_values_by_metric_name = {}
            for row in rows:
                group = {}
                metric_values = {}

                for i in range(len(row)):
                    metric_name = sql_metric.metric_names[i] if sql_metric.metric_names is not None else description[i][
                        0]
                    metric_value = row[i]
                    if metric_name.lower() in group_fields_lower:
                        group[metric_name] = metric_value
                    else:
                        metric_values[metric_name] = metric_value

                if not group:
                    logger.error(f'None of the declared group_fields were found in '
                                 f'result: {sql_metric.group_fields}. Skipping result.')
                else:
                    for metric_name in metric_values:
                        metric_value = metric_values[metric_name]
                        if metric_name not in group_values_by_metric_name:
                            group_values_by_metric_name[metric_name] = []
                        group_values = group_values_by_metric_name[metric_name]
                        group_values.append(GroupValue(group=group, value=metric_value))
                        logger.debug(f'SQL metric {sql_metric.title} {metric_name} {group} -> {metric_value}')

                    sql_metric_tests = sql_metric.tests
                    test_variables = self._get_test_variables(scan_column)
                    test_variables.update(metric_values)
                    sql_metric_test_results = self._execute_tests(sql_metric_tests, test_variables, group)
                    test_results.extend(sql_metric_test_results)

            column_name = scan_column.column_name if scan_column is not None else None
            for metric_name in group_values_by_metric_name:
                group_values = group_values_by_metric_name[metric_name]
                measurement = Measurement(metric=metric_name, group_values=group_values, column_name=column_name)
                self._log_and_append_query_measurement(measurements, measurement)

            self._flush_measurements(measurements)
            self._flush_test_results(test_results)
        except Exception as e:
            self.scan_result.add_error(ScanError(f'Exception during sql metric groups query {resolved_sql}', e))

    def _send_failed_rows_custom_metric(self,
                                        metric_name: str,
                                        column_name: str,
                                        sql: str,
                                        metric_id: str):
        try:
            if self.soda_server_client:

                logger.debug(f'Sending failed rows for {metric_id} to Soda Cloud')
                with tempfile.TemporaryFile() as temp_file:

                    failed_limit = 5
                    if self.scan_yml.samples_yml is not None:
                        failed_limit = self.scan_yml.samples_yml.failed_limit or 5
                    sql += f'\nLIMIT {failed_limit}'

                    stored_failed_rows, sample_columns, total_failed_rows = \
                        self.sampler.save_sample_to_local_file_with_limit(sql, temp_file, failed_limit)

                    if stored_failed_rows > 0:
                        temp_file_size_in_bytes = temp_file.tell()
                        temp_file.seek(0)

                        file_path = self.sampler.create_file_path_failed_rows_sql_metric(column_name=column_name,
                                                                                         metric_name=metric_name)
                        file_id = self.soda_server_client.scan_upload(self.scan_reference,
                                                                      file_path,
                                                                      temp_file,
                                                                      temp_file_size_in_bytes)

                        self.soda_server_client.scan_file(
                            scan_reference=self.scan_reference,
                            sample_type='failedRowsSample',
                            stored=int(stored_failed_rows),
                            total=int(total_failed_rows),
                            source_columns=sample_columns,
                            file_id=file_id,
                            column_name=column_name,
                            test_ids=[],
                            sql_metric_name=metric_name,
                            custom_metric_id=metric_id)
                        logger.debug(
                            f'Sent failed rows for id: {metric_id} ({stored_failed_rows}/{total_failed_rows}) to Soda Cloud')
                    else:
                        logger.debug(f'No failed rows for custom metric ({metric_name} with id {metric_id})')
            else:
                failed_rows_tuples, description = self.warehouse.sql_fetchall_description(sql=sql)
                if len(failed_rows_tuples) > 0:
                    table_text = self._table_to_text(failed_rows_tuples, description)
                    logger.debug(f'Failed rows for sql metric sql {metric_name}:\n' + table_text)
                else:
                    logger.debug(f'No failed rows for custom metric ({metric_name} with id {metric_id})')
        except Exception as e:
            logger.exception(f'Could not perform sql metric failed rows \n{sql}', e)
            self.scan_result.add_error(ScanError(f'Exception during sql metric failed rows query {sql}', e))

    def _run_sql_metric_failed_rows(self,
                                    sql_metric_yml: SqlMetricYml,
                                    resolved_sql: str,
                                    scan_column: Optional[ScanColumn] = None):

        try:
            if self.soda_server_client:

                logger.debug(f'Sending failed rows for sql metric {sql_metric_yml.name} to Soda Cloud')
                with tempfile.TemporaryFile() as temp_file:

                    failed_limit = self.scan_yml.get_sql_metric_failed_rows_limit(sql_metric_yml)

                    stored_failed_rows, sample_columns, total_failed_rows = \
                        self.sampler.save_sample_to_local_file_with_limit(resolved_sql, temp_file, failed_limit)

                    column_name = scan_column.column_name if scan_column else None
                    measurement = Measurement(metric=sql_metric_yml.name, value=total_failed_rows,
                                              column_name=column_name)

                    measurements = []
                    self._log_and_append_query_measurement(measurements, measurement)
                    self._flush_measurements(measurements)

                    test = sql_metric_yml.tests[0]

                    test_variables = {
                        sql_metric_yml.name: total_failed_rows
                    }

                    test_result = self.evaluate_test(test, test_variables)
                    self._flush_test_results([test_result])

                    if stored_failed_rows > 0:
                        temp_file_size_in_bytes = temp_file.tell()
                        temp_file.seek(0)

                        file_path = self.sampler.create_file_path_failed_rows_sql_metric(
                            column_name=sql_metric_yml.column_name,
                            metric_name=sql_metric_yml.name)

                        file_id = self.soda_server_client.scan_upload(self.scan_reference,
                                                                      file_path,
                                                                      temp_file,
                                                                      temp_file_size_in_bytes)

                        self.soda_server_client.scan_file(
                            scan_reference=self.scan_reference,
                            sample_type='failedRowsSample',
                            stored=int(stored_failed_rows),
                            total=int(total_failed_rows),
                            source_columns=sample_columns,
                            file_id=file_id,
                            column_name=sql_metric_yml.column_name,
                            test_ids=[test.id],
                            sql_metric_name=sql_metric_yml.name)
                        logger.debug(
                            f'Sent failed rows for sql metric ({stored_failed_rows}/{total_failed_rows}) to Soda Cloud')
                    else:
                        logger.debug(f'No failed rows for sql metric ({sql_metric_yml.name})')
            else:
                failed_rows_tuples, description = self.warehouse.sql_fetchall_description(sql=resolved_sql)
                if len(failed_rows_tuples) > 0:
                    table_text = self._table_to_text(failed_rows_tuples, description)
                    logger.debug(f'Failed rows for sql metric sql {sql_metric_yml.name}:\n' + table_text)
                else:
                    logger.debug(f'No failed rows for sql metric sql {sql_metric_yml.name}')
        except Exception as e:
            logger.exception(f'Could not perform sql metric failed rows \n{resolved_sql}', e)
            self.scan_result.add_error(ScanError(f'Exception during sql metric failed rows query {resolved_sql}', e))

    def _get_historic_metric_variables(self, scan_column: ScanColumn):
        historic_metrics: List[HistoricMetricYml] = scan_column.get_historic_metrics()
        historic_variables = {}
        if historic_metrics:
            logger.info('Fetching historic metrics')
            if self.soda_server_client:
                metrics = []
                for hmc in historic_metrics:
                    metric = {
                        'key': hmc.name,
                        'metricType': hmc.metric,
                    }
                    if hmc.type == 'prev':
                        metric['previousCompare'] = {'count': hmc.count}
                    else:
                        metric['aggregatedCompare'] = {
                            'type': hmc.type,
                            'count': hmc.count
                        }
                    metric['columnName'] = scan_column.column_name
                    metrics.append(metric)
                results = self.soda_server_client.historic_metrics(self.warehouse, self.scan_yml.table_name, metrics)
                measurements = results['measurements']
                errors = []
                for name, val in measurements.items():
                    if 'value' in val:
                        historic_variables[name] = val['value']
                    elif 'error' in val:
                        errors.append(f"Error in fetching historical metric {name}: {val['error']['message']}")
                if errors:
                    raise RuntimeError('\n'.join(errors))
            else:
                raise RuntimeError("Can't use historic metrics in tests without a connection to Soda Cloud")
        return historic_variables

    def _get_test_variables(self, scan_column: Optional[ScanColumn] = None):
        column_name_lower = scan_column.column_name_lower if scan_column is not None else None
        return {
            measurement.metric: measurement.value
            for measurement in self.scan_result.measurements
            if measurement.column_name is None or measurement.column_name.lower() == column_name_lower
        }

    def _run_table_tests(self):
        test_variables = self._get_test_variables()

        table_tests = self.scan_yml.tests
        table_test_results = self._execute_tests(table_tests, test_variables)
        self._flush_test_results(table_test_results)

    def _run_column_tests(self):
        test_results = []
        for column_name_lower, scan_column in self.scan_columns.items():
            column_tests = scan_column.get_tests()
            test_variables = self._get_test_variables(scan_column)
            historic_metric_variables = self._get_historic_metric_variables(scan_column)
            # Shallow merge two variable dicts
            test_variables = {**test_variables, **historic_metric_variables}
            column_test_results = self._execute_tests(column_tests, test_variables)
            test_results.extend(column_test_results)
        self._flush_test_results(test_results)

    def _execute_tests(self, tests: List[Test], variables, group_values: Optional[dict] = None):
        test_results = []
        if tests:
            for test in tests:
                test_result = self.evaluate_test(test, variables, group_values)
                test_results.append(test_result)
        return test_results

    def _process_samples(self):
        try:
            for measurement in self.scan_result.measurements:
                if (measurement.metric in [Metric.ROW_COUNT, Metric.MISSING_COUNT, Metric.INVALID_COUNT,
                                           Metric.VALUES_COUNT, Metric.VALID_COUNT] and measurement.value > 0):
                    samples_yml = self.scan_yml.get_sample_yml(measurement)
                    if samples_yml:
                        sample_details = self.sampler.get_samples(samples_yml, measurement,
                                                                  self.scan_result.test_results)
                        if self.failed_rows_processor:
                            self.failed_rows_processor.process(sample_details)
                        if self.soda_server_client:
                            self.sampler.send_samples_to_soda_cloud(sample_details)

        except Exception as e:
            logger.exception(f'Could not process samples: {e}')
            self.scan_result.add_error(SodaCloudScanError('Could not process samples', e))

    @classmethod
    def _log_and_append_query_measurement(cls, measurements: List[Measurement], measurement: Measurement):
        """
        Convenience method to log a measurement and append it to the given list of measurements
        Logging will indicate it is the result of a query
        """
        cls._log_measurement(measurement)
        measurements.append(measurement)

    @classmethod
    def _log_and_append_derived_measurements(
        cls,
        measurements: List[Measurement],
        derived_measurements: List[Measurement]):
        """
        Convenience method to log a list of derived measurements and append them to the given list of measurements.
        Logging will indicate it is a derived measurement
        """
        for derived_measurement in derived_measurements:
            cls._log_measurement(derived_measurement, is_derived=True)
            measurements.append(derived_measurement)

    @classmethod
    def _log_measurement(cls, measurement, is_derived: bool = False):
        measurement_type = "Derived" if is_derived else "Query"
        logger.debug(f'{measurement_type} measurement: {measurement}')

    def _flush_measurements(self, measurements: List[Measurement]):
        """
        Adds the measurements to the scan result and sends the measurements to the Soda Server if that's configured
        """
        self.scan_result.measurements.extend(measurements)
        if self.soda_server_client and measurements:
            measurement_jsons = [measurement.to_dict() for measurement in measurements]
            try:
                self.soda_server_client.scan_measurements(self.scan_reference, measurement_jsons)
            except Exception as e:
                logger.error(f'Soda Cloud error: Could not send measurements: {e}')
                self.scan_result.add_error(SodaCloudScanError('Could not send measurements', e))

    def _flush_test_results(self, test_results: List[TestResult]):
        """
        Adds the test_results to the scan result and sends the measurements to the Soda Server if that's configured
        """
        if test_results:
            self.scan_result.add_test_results(test_results)

            if self.soda_server_client:
                test_result_jsons = [test_result.to_dict() for test_result in test_results]
                try:
                    self.soda_server_client.scan_test_results(self.scan_reference, test_result_jsons)
                except Exception as e:
                    logger.error(f'Soda Cloud error: Could not send test results: {e}')
                    self.scan_result.add_error(SodaCloudScanError('Could not send test results', e))

    def _ensure_scan_reference(self):
        if self.soda_server_client and not self.scan_reference:
            try:
                self.start_scan_response = self.soda_server_client.scan_start(
                    self.warehouse,
                    self.scan_yml,
                    self.time,
                    origin=os.environ.get('SODA_SCAN_ORIGIN', 'external'))
                self.scan_reference = self.start_scan_response['scanReference']
                if 'disableCollectingWarehouseData' in self.start_scan_response:
                    self.disable_sample_collection = self.start_scan_response['disableCollectingWarehouseData']

            except Exception as e:
                logger.error(f'Soda Cloud error: Could not start scan: {e}')
                logger.error(f'Skipping subsequent Soda Cloud communication but continuing the scan')
                self.scan_result.add_error(SodaCloudScanError('Could not start scan', e))
                self.soda_server_client = None

    def evaluate_test(self, test: Test, test_variables, group_values: Optional[dict] = None) -> TestResult:
        test_result = test.evaluate(test_variables, group_values, self.variables)
        if test_result.error:
            self.scan_result.add_error(TestExecutionScanError(
                message=f'Test "{test_result.test.expression}" failed',
                exception=test_result.error,
                test=test_result.test
            ))
        return test_result

    def _table_to_text(self, failed_row_tuples: tuple, description):
        columns = [col[0] for col in description]
        widths = [len(col) for col in columns]
        tavnit = '|'
        separator = '+'

        for failed_row_tuple in failed_row_tuples:
            failed_row_tuple = tuple(str(cell) for cell in failed_row_tuple)
            for i in range(0, len(failed_row_tuple)):
                widths[i] = max(widths[i], len(failed_row_tuple[i]))

        for w in widths:
            tavnit += " %-" + "%ss |" % (w,)
            separator += '-' * w + '--+'

        table_text = ''
        table_text += separator + '\n'
        table_text += tavnit % tuple(columns) + '\n'
        table_text += separator + '\n'
        for row in failed_row_tuples:
            table_text += tavnit % row + '\n'
        table_text += separator
        return table_text



#!/usr/bin/env python
import sys
import pathlib
from setuptools import setup, find_namespace_packages

if sys.version_info < (3, 7):
    print('Error: Soda SQL requires at least Python 3.7')
    print('Error: Please upgrade your Python version to 3.7 or later')
    sys.exit(1)

package_name = "soda-sql-core"
# Managed by tbump - don't change manually
# And we can't have nice semver (<major>.<minor>.<patch>-<pre-release>-<build>)
# like "-alpha-1" as long as this is open >> https://github.com/pypa/setuptools/issues/2181
package_version = '2.1.0'
description = "Soda SQL Core"

long_description = (pathlib.Path(__file__).parent / "README.md").read_text()

requires = [
    'Jinja2>=2.11.3, <3.0',
    'click>=7.1.2, <8.0',
    'cryptography==3.3.2',
    'pyyaml>=5.4.1, <6.0',
    'requests>=2.23.0, <3.0',
    'Deprecated>=1.2.13, <1.3',
    'opentelemetry-exporter-otlp-proto-http>=1.6.2,<1.7',
]
# TODO Fix the params
# TODO Add a warning that installing core doesn't give any warehouse functionality
setup(
    name=package_name,
    version=package_version,
    author="Tom Baeyens",
    author_email="tom@soda.io",
    description="Soda SQL library & CLI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_namespace_packages(include=["sodasql*"]),
    install_requires=requires,
    entry_points={
        "console_scripts":
            [
                "soda=sodasql.cli.cli:main"
            ]
    },
    classifiers=[
        'Development Status :: 4 - Beta',

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',

        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires=">=3.7",
)



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import re
from typing import List, Optional, Set

from jinja2 import Template

from sodasql.scan.file_system import FileSystemSingleton
from sodasql.scan.historic_metric_yml import HistoricMetricYml
from sodasql.scan.metric import Metric
from sodasql.scan.missing import Missing
from sodasql.scan.parser import Parser
from sodasql.scan.samples_yml import SamplesYml
from sodasql.scan.scan_yml import ScanYml
from sodasql.scan.scan_yml_column import ScanYmlColumn
from sodasql.scan.sql_metric_yml import SqlMetricYml
from sodasql.scan.test import Test
from sodasql.scan.validity import Validity
from sodasql.common.yaml_helper import YamlHelper

KEY_TABLE_NAME = 'table_name'
KEY_METRICS = 'metrics'
KEY_HISTORIC_METRICS = 'historic_metrics'
KEY_METRIC_GROUPS = 'metric_groups'
KEY_SQL_METRICS = 'sql_metrics'
KEY_TESTS = 'tests'
KEY_COLUMNS = 'columns'
KEY_EXCLUDED_COLUMNS = 'excluded_columns'
KEY_MINS_MAXS_LIMIT = 'mins_maxs_limit'
KEY_FREQUENT_VALUES_LIMIT = 'frequent_values_limit'
KEY_SAMPLE_PERCENTAGE = 'sample_percentage'
KEY_SAMPLE_METHOD = 'sample_method'
KEY_FILTER = 'filter'
KEY_SAMPLES = 'samples'

VALID_SCAN_YML_KEYS = [KEY_TABLE_NAME, KEY_METRICS, KEY_METRIC_GROUPS, KEY_SQL_METRICS,
                       KEY_TESTS, KEY_COLUMNS, KEY_MINS_MAXS_LIMIT, KEY_FREQUENT_VALUES_LIMIT,
                       KEY_SAMPLE_PERCENTAGE, KEY_SAMPLE_METHOD, KEY_FILTER, KEY_SAMPLES, KEY_EXCLUDED_COLUMNS]

COLUMN_KEY_METRICS = KEY_METRICS
COLUMN_KEY_METRIC_GROUPS = KEY_METRIC_GROUPS
COLUMN_KEY_SQL_METRICS = KEY_SQL_METRICS
COLUMN_KEY_TESTS = KEY_TESTS
COLUMN_KEY_SAMPLES = KEY_SAMPLES
COLUMN_KEY_HISTORIC_METRICS = KEY_HISTORIC_METRICS

COLUMN_KEY_MISSING_VALUES = 'missing_values'
COLUMN_KEY_MISSING_FORMAT = 'missing_format'
COLUMN_KEY_MISSING_REGEX = 'missing_regex'

COLUMN_KEY_VALID_FORMAT = 'valid_format'
COLUMN_KEY_VALID_REGEX = 'valid_regex'
COLUMN_KEY_VALID_VALUES = 'valid_values'
COLUMN_KEY_VALID_MIN = 'valid_min'
COLUMN_KEY_VALID_MAX = 'valid_max'
COLUMN_KEY_VALID_MIN_LENGTH = 'valid_min_length'
COLUMN_KEY_VALID_MAX_LENGTH = 'valid_max_length'

COLUMN_MISSING_KEYS = [
    COLUMN_KEY_MISSING_VALUES,
    COLUMN_KEY_MISSING_FORMAT,
    COLUMN_KEY_MISSING_REGEX]

COLUMN_VALID_KEYS = [
    COLUMN_KEY_VALID_FORMAT,
    COLUMN_KEY_VALID_REGEX,
    COLUMN_KEY_VALID_VALUES,
    COLUMN_KEY_VALID_MIN,
    COLUMN_KEY_VALID_MAX,
    COLUMN_KEY_VALID_MIN_LENGTH,
    COLUMN_KEY_VALID_MAX_LENGTH]

VALID_COLUMN_KEYS = COLUMN_MISSING_KEYS + COLUMN_VALID_KEYS + [
    COLUMN_KEY_METRICS,
    COLUMN_KEY_METRIC_GROUPS,
    COLUMN_KEY_SQL_METRICS,
    COLUMN_KEY_TESTS,
    COLUMN_KEY_SAMPLES,
    COLUMN_KEY_HISTORIC_METRICS]

SAMPLES_KEY_DATASET_LIMIT = 'table_limit'
SAMPLES_KEY_DATASET_TABLESAMPLE = 'table_tablesample'
SAMPLES_KEY_FAILED_LIMIT = 'failed_limit'
SAMPLES_KEY_FAILED_TABLESAMPLE = 'failed_tablesample'
SAMPLES_KEY_PASSED_LIMIT = 'passed_limit'
SAMPLES_KEY_PASSED_TABLESAMPLE = 'passed_tablesample'

SQL_METRIC_KEY_NAME = 'name'
SQL_METRIC_KEY_TITLE = 'title'
SQL_METRIC_KEY_SQL = 'sql'
SQL_METRIC_KEY_TYPE = 'type'
SQL_METRIC_KEY_SQL_FILE = 'sql_file'
SQL_METRIC_KEY_METRIC_NAMES = 'metric_names'
SQL_METRIC_KEY_TESTS = KEY_TESTS
SQL_METRIC_KEY_GROUP_FIELDS = 'group_fields'
SQL_METRIC_KEY_FAILED_LIMIT = SAMPLES_KEY_FAILED_LIMIT

VALID_SQL_METRIC_KEYS = [SQL_METRIC_KEY_SQL, SQL_METRIC_KEY_METRIC_NAMES, SQL_METRIC_KEY_TESTS,
                         SQL_METRIC_KEY_SQL_FILE, SQL_METRIC_KEY_GROUP_FIELDS, SQL_METRIC_KEY_FAILED_LIMIT]


class ScanYmlParser(Parser):

    def __init__(self,
                 scan_yml_dict: dict,
                 scan_yml_path: str = 'scan'):
        super().__init__(description=scan_yml_path)

        self.scan_yaml_path = scan_yml_path

        self.scan_yml = ScanYml()

        self._push_context(scan_yml_dict, self.description)

        table_name = self.get_str_required(KEY_TABLE_NAME)
        self.scan_yml.table_name = table_name
        self.scan_yml.metrics = self.parse_metrics()
        self.scan_yml.excluded_columns = self.get_list_optional(KEY_EXCLUDED_COLUMNS)

        self.scan_yml.sql_metric_ymls = self.parse_sql_metric_ymls(KEY_SQL_METRICS)

        self.scan_yml.tests = self.parse_tests(scan_yml_dict, KEY_TESTS, context_table_name=table_name)
        self.scan_yml.columns = self.parse_columns(self.scan_yml)
        self.scan_yml.mins_maxs_limit = self.get_int_optional(KEY_MINS_MAXS_LIMIT, 5)
        self.scan_yml.frequent_values_limit = self.get_int_optional(KEY_FREQUENT_VALUES_LIMIT, 5)
        self.scan_yml.samples_yml = self.parse_samples_yml(KEY_SAMPLES)

        # TODO change the next 2 properties to filter_tablesample (similar to samples.table_tablesample)
        # afaict, they are not docced and can be changed.
        self.scan_yml.sample_percentage = self.get_float_optional(KEY_SAMPLE_PERCENTAGE)
        self.scan_yml.sample_method = self.get_str_optional(KEY_SAMPLE_METHOD, 'SYSTEM').upper()

        filter = self.get_str_optional(KEY_FILTER)
        if filter:
            try:
                self.scan_yml.filter = filter
                self.scan_yml.filter_template = Template(filter)
            except Exception as e:
                self.error(f"Couldn't parse filter '{filter}': {str(e)}", KEY_FILTER)

        self.check_invalid_keys(VALID_SCAN_YML_KEYS)

    def parse_historic_metrics(self):
        historic_metrics_yml = self.get_list_optional(KEY_HISTORIC_METRICS, [])
        historic_metrics: List[HistoricMetricYml] = []
        self._push_context(historic_metrics, KEY_HISTORIC_METRICS)
        for hmc in historic_metrics_yml:
            historic_metrics.append(
                HistoricMetricYml(name=hmc['name'],
                                  type=hmc['type'],
                                  metric=hmc['metric'],
                                  count=1 if hmc['count'] is None else hmc['count'])
            )

        self._pop_context()
        return historic_metrics

    def parse_metrics(self):
        metrics: Set[str] = set(self.get_list_optional(KEY_METRICS, []))
        metrics_groups: Set[str] = set(self.get_list_optional(KEY_METRIC_GROUPS, []))

        self._push_context(metrics, KEY_METRICS)

        # Add special case for all
        if Metric.METRIC_GROUP_ALL in metrics_groups:
            metrics.update(Metric.METRIC_TYPES)
        else:
            for metric_group_name in Metric.DERIVED_METRIC_GROUPS:
                group_metrics = Metric.METRIC_GROUPS[metric_group_name]
                for metric in metrics:
                    if metric in group_metrics:
                        metrics_groups.add(metric_group_name)

            if Metric.METRIC_GROUP_VALIDITY in metrics_groups:
                metrics_groups.add(Metric.METRIC_GROUP_MISSING)

            if Metric.METRIC_GROUP_MISSING in metrics_groups:
                metrics.add(Metric.ROW_COUNT)

            for metric_group_name in metrics_groups:
                if metric_group_name in Metric.METRIC_GROUPS:
                    for group_metric in Metric.METRIC_GROUPS[metric_group_name]:
                        metrics.add(group_metric)
                else:
                    self.warning(f'Invalid metric_group {metric_group_name}')

            if Metric.HISTOGRAM in metrics:
                metrics.add(Metric.MIN)
                metrics.add(Metric.MAX)

        self.check_invalid_keys(Metric.METRIC_TYPES)

        self._pop_context()

        return metrics

    def ensure_metric(self,
                      metrics: Set[str],
                      metric: str,
                      dependent_metric: str,
                      column_name: str = None):
        if metric not in metrics:
            metrics.add(metric)
            column_message = f' on column {column_name}' if column_name else ''
            self.info(f'Added metric {metric} as dependency of {dependent_metric}{column_message}')

    def is_metric_category_enabled(self, metrics: Set[str], category: str, category_metrics: List[str]):
        if category in metrics:
            return True
        for category_metric in category_metrics:
            if category_metric in metrics:
                return True
        return False

    def resolve_category(self,
                         metrics: Set[str],
                         metrics_groups: List[str],
                         category: str,
                         category_metrics: List[str],
                         column_name: str = None):
        if self.is_metric_category_enabled(metrics, category, category_metrics):
            if category in metrics:
                metrics.remove(category)
            for category_metric in category_metrics:
                self.ensure_metric(metrics, category_metric, category, column_name)

    def remove_metric(self, metrics: Set[str], metric: str):
        """Returns True if the metric was removed from the set"""
        if metric in metrics:
            metrics.remove(metric)
            return True
        else:
            return False

    def parse_columns(self, scan_configuration: ScanYml) -> dict:
        columns_dict = self.get_dict_optional(KEY_COLUMNS, {})
        self._push_context(columns_dict, KEY_COLUMNS)

        scan_configuration_columns = {}

        for column_name in columns_dict:
            column_dict = columns_dict.get(column_name)
            if not isinstance(column_dict, dict):
                self.error(f'Column {column_name} should be an object, not a {type(column_dict)}', KEY_COLUMNS)
            else:
                self._push_context(column_dict, column_name)

                metrics: Set[str] = self.parse_metrics()

                if self.remove_metric(metrics, Metric.ROW_COUNT):
                    self.ensure_metric(scan_configuration.metrics, Metric.ROW_COUNT,
                                       f'{Metric.ROW_COUNT} {column_name}')

                missing = None
                if any(cfg in column_dict.keys() for cfg in COLUMN_MISSING_KEYS):
                    missing = Missing()
                    missing.values = column_dict.get(COLUMN_KEY_MISSING_VALUES)
                    missing.format = column_dict.get(COLUMN_KEY_MISSING_FORMAT)
                    missing.regex = column_dict.get(COLUMN_KEY_MISSING_REGEX)

                validity = None

                if any(cfg in column_dict.keys() for cfg in COLUMN_VALID_KEYS):
                    validity = Validity()
                    validity.format = column_dict.get(COLUMN_KEY_VALID_FORMAT)
                    if validity.format is not None and Validity.FORMATS.get(validity.format) is None:
                        self.warning(f'Invalid {column_name}.{COLUMN_KEY_VALID_FORMAT}: {validity.format}')
                    validity.regex = column_dict.get(COLUMN_KEY_VALID_REGEX)
                    validity.values = YamlHelper.validate_list_value(column_name, COLUMN_KEY_VALID_VALUES,
                                                                     column_dict.get(COLUMN_KEY_VALID_VALUES))
                    validity.min = YamlHelper.validate_numeric_value(column_name, COLUMN_KEY_VALID_MIN,
                                                                     column_dict.get(COLUMN_KEY_VALID_MIN))
                    validity.max = YamlHelper.validate_numeric_value(column_name, COLUMN_KEY_VALID_MAX,
                                                                     column_dict.get(COLUMN_KEY_VALID_MAX))
                    validity.min_length = column_dict.get(COLUMN_KEY_VALID_MIN_LENGTH)
                    validity.max_length = column_dict.get(COLUMN_KEY_VALID_MAX_LENGTH)

                sql_metric_ymls = self.parse_sql_metric_ymls(metrics_key=COLUMN_KEY_SQL_METRICS,
                                                             column_name=column_name)

                tests = self.parse_tests(column_dict,
                                         COLUMN_KEY_TESTS,
                                         context_table_name=scan_configuration.table_name,
                                         context_column_name=column_name)

                samples_yml = self.parse_samples_yml(COLUMN_KEY_SAMPLES)
                if samples_yml and samples_yml.table_limit:
                    self.warning(f"Invalid column samples key 'table_limit'")
                if samples_yml and samples_yml.table_tablesample:
                    self.warning(f"Invalid column samples key 'table_tablesample'")

                self.check_invalid_keys(VALID_COLUMN_KEYS)

                historic_metrics = self.parse_historic_metrics()

                column_name_lower = column_name.lower()
                scan_configuration_columns[column_name_lower] = ScanYmlColumn(
                    metrics=metrics,
                    sql_metric_ymls=sql_metric_ymls,
                    missing=missing,
                    validity=validity,
                    tests=tests,
                    samples_yml=samples_yml,
                    historic_metrics=historic_metrics
                )

                self._pop_context()

        self._pop_context()

        return scan_configuration_columns

    def parse_sql_metric_ymls(self, metrics_key: str, column_name: Optional[str] = None) -> List[SqlMetricYml]:
        sql_metrics_dicts: List[dict] = self.get_list_optional(metrics_key, [])
        if isinstance(sql_metrics_dicts, list):
            self._push_context(sql_metrics_dicts, metrics_key)
            try:
                sql_metric_ymls = []
                for i in range(len(sql_metrics_dicts)):
                    sql_metric_dict = sql_metrics_dicts[i]
                    if isinstance(sql_metric_dict, dict):
                        metric_type = sql_metric_dict.get(SQL_METRIC_KEY_TYPE, 'numeric')
                        metric_group_fields = sql_metric_dict.get(SQL_METRIC_KEY_GROUP_FIELDS)
                        if metric_group_fields:
                            metric_type = 'numeric_groups'

                        self._push_context(object=sql_metric_dict, name=i)
                        try:
                            sql_metric_yml = None

                            if metric_type == 'failed_rows':
                                sql_metric_yml = self.parse_sql_metric_failed_rows(
                                    metric_type,
                                    sql_metric_index=i,
                                    column_name=column_name)

                            elif metric_type.startswith('numeric'):
                                sql_metric_yml = self.parse_sql_metric(
                                    sql_metric_dict=sql_metric_dict,
                                    sql_metric_type=metric_type,
                                    sql_metric_index=i,
                                    column_name=column_name)
                            else:
                                self.error(f'Unknown sql_metric type {metric_type}')

                            if sql_metric_yml:
                                sql_metric_ymls.append(sql_metric_yml)

                        finally:
                            self._pop_context()
                    else:
                        self.error(f'Invalid YAML: SQL metric {i} was {type(sql_metric_dict)}, expected object')

                return sql_metric_ymls
            finally:
                self._pop_context()
        elif sql_metrics_dicts is not None:
            self.error(
                f'Invalid YAML structure near {metrics_key}: Expected list of SQL metrics, but was {type(sql_metrics_dicts)}',
                metrics_key)

    def parse_sql_metric(self,
                         sql_metric_dict,
                         sql_metric_type: str,
                         sql_metric_index: int,
                         column_name: Optional[str] = None) -> Optional[SqlMetricYml]:
        sql_metric_name = self.get_str_optional(SQL_METRIC_KEY_NAME)
        metric_names = self.get_list_optional(SQL_METRIC_KEY_METRIC_NAMES)
        group_fields = self.get_list_optional(SQL_METRIC_KEY_GROUP_FIELDS)
        sql = self.get_str_optional(SQL_METRIC_KEY_SQL)
        sql_file = self.get_str_optional(SQL_METRIC_KEY_SQL_FILE)

        if not sql and not sql_file:
            self.error('No sql nor sql_file specified in SQL metric')
        elif sql_file:
            file_system = FileSystemSingleton.INSTANCE
            sql_file_path = file_system.join(file_system.dirname(self.scan_yaml_path), sql_file)
            sql = file_system.file_read_as_str(sql_file_path)

        tests = self.parse_tests(
            sql_metric_dict,
            SQL_METRIC_KEY_TESTS,
            context_table_name=self.scan_yml.table_name,
            context_column_name=column_name,
            context_sql_metric_name=sql_metric_name,
            context_sql_metric_index=sql_metric_index)

        sql_metric_title = ((column_name + '.' if column_name else '') +
                            (sql_metric_name if sql_metric_name else f'sql_metric[{sql_metric_index}]'))

        sql_metric_yml: SqlMetricYml = SqlMetricYml(type=sql_metric_type,
                                                    name=None,
                                                    title=sql_metric_title,
                                                    sql=sql,
                                                    index=sql_metric_index,
                                                    column_name=column_name,
                                                    metric_names=metric_names,
                                                    group_fields=group_fields,
                                                    tests=tests)

        self.check_invalid_keys(VALID_SQL_METRIC_KEYS)
        return sql_metric_yml

    def parse_sql_metric_failed_rows(self,
                                     sql_metric_type: str,
                                     sql_metric_index: int,
                                     column_name: str) -> Optional[SqlMetricYml]:

        sql_metric_name = self.get_str_required(SQL_METRIC_KEY_NAME)
        if sql_metric_name is None:
            return None

        # Only allow for valid python identifiers as the metric names in failed rows
        # TODO make this consistent with the other sql metric names
        if not re.match(r"^[^\d\W]\w*\Z$", sql_metric_name):
            self.error(f'Invalid metric identifier {sql_metric_name}', SQL_METRIC_KEY_NAME)
            return None

        test_name = sql_metric_name
        test_expression = f'{test_name} == 0'
        test_id = self.create_test_id(test_expression, test_name, sql_metric_index, column_name, sql_metric_name,
                                      sql_metric_index)
        test_title = self.create_test_title(test_expression, test_name, sql_metric_index, column_name, sql_metric_name,
                                            sql_metric_index)

        test = Test(id=test_id,
                    title=test_title,
                    expression=test_expression,
                    metrics=[sql_metric_name],
                    column=column_name)

        return SqlMetricYml(
            type=sql_metric_type,
            name=sql_metric_name,
            title=self.get_str_optional(SQL_METRIC_KEY_TITLE),
            sql=self.get_str_optional(SQL_METRIC_KEY_SQL),
            column_name=column_name,
            index=sql_metric_index,
            failed_limit=self.get_int_optional(SQL_METRIC_KEY_FAILED_LIMIT),
            tests=[test])

    def parse_samples_yml(self, samples_key: str) -> Optional[SamplesYml]:
        samples = self.get_dict_optional(samples_key)
        if samples is not None:
            self._push_context(samples, samples_key)
            try:
                samples_yml = SamplesYml(
                    table_limit=self.get_int_optional(SAMPLES_KEY_DATASET_LIMIT),
                    table_tablesample=self.get_str_optional(SAMPLES_KEY_DATASET_TABLESAMPLE),
                    failed_limit=self.get_int_optional(SAMPLES_KEY_FAILED_LIMIT),
                    failed_tablesample=self.get_str_optional(SAMPLES_KEY_FAILED_TABLESAMPLE),
                    passed_limit=self.get_int_optional(SAMPLES_KEY_PASSED_LIMIT),
                    passed_tablesample=self.get_str_optional(SAMPLES_KEY_PASSED_TABLESAMPLE)
                )
                return samples_yml
            finally:
                self._pop_context()
        pass
